##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for feed with flash.
Authors: Andrew Lee
"""

import pytest
from pyomo.environ import (ConcreteModel,
                           TerminationCondition,
                           SolverStatus,
                           value)
from idaes.core import FlowsheetBlock, MaterialBalanceType
from idaes.generic_models.unit_models.feed_flash import FeedFlash, FlashType
from idaes.generic_models.properties import iapws95
from idaes.generic_models.properties.activity_coeff_models.BTX_activity_coeff_VLE \
    import BTXParameterBlock
from idaes.core.util.model_statistics import (degrees_of_freedom,
                                              number_variables,
                                              number_total_constraints,
                                              fixed_variables_set,
                                              activated_constraints_set,
                                              number_unused_variables)
from idaes.core.util.testing import (PhysicalParameterTestBlock,
                                     initialization_tester)
from idaes.core.util import get_solver
from pyomo.util.check_units import assert_units_consistent


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


# -----------------------------------------------------------------------------
@pytest.mark.unit
def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.properties = PhysicalParameterTestBlock()

    m.fs.unit = FeedFlash(default={"property_package": m.fs.properties})

    # Check unit config arguments
    assert len(m.fs.unit.config) == 6

    assert not m.fs.unit.config.dynamic
    assert not m.fs.unit.config.has_holdup
    assert m.fs.unit.config.material_balance_type == \
        MaterialBalanceType.useDefault
    assert m.fs.unit.config.flash_type == FlashType.isothermal
    assert m.fs.unit.config.property_package is m.fs.properties


# -----------------------------------------------------------------------------
class TestBTXIdeal(object):
    @pytest.fixture(scope="class")
    def btx(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = BTXParameterBlock(default={"valid_phase":
                                                     ('Liq', 'Vap'),
                                                     "activity_coeff_model":
                                                     "Ideal"})

        m.fs.unit = FeedFlash(default={"property_package": m.fs.properties})

        m.fs.unit.flow_mol.fix(1)
        m.fs.unit.temperature.fix(368)
        m.fs.unit.pressure.fix(101325)
        m.fs.unit.mole_frac_comp[0, "benzene"].fix(0.5)
        m.fs.unit.mole_frac_comp[0, "toluene"].fix(0.5)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, btx):
        assert hasattr(btx.fs.unit, "flow_mol")
        assert hasattr(btx.fs.unit, "mole_frac_comp")
        assert hasattr(btx.fs.unit, "temperature")
        assert hasattr(btx.fs.unit, "pressure")

        assert hasattr(btx.fs.unit, "outlet")
        assert len(btx.fs.unit.outlet.vars) == 4
        assert hasattr(btx.fs.unit.outlet, "flow_mol")
        assert hasattr(btx.fs.unit.outlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.outlet, "temperature")
        assert hasattr(btx.fs.unit.outlet, "pressure")

        assert hasattr(btx.fs.unit, "isothermal")

        assert number_variables(btx) == 34
        assert number_total_constraints(btx) == 29
        assert number_unused_variables(btx) == 0

    @pytest.mark.component
    def test_units(self, btx):
        assert_units_consistent(btx)

    @pytest.mark.unit
    def test_dof(self, btx):
        assert degrees_of_freedom(btx) == 0

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, btx):
        initialization_tester(btx)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, btx):
        results = solver.solve(btx)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, btx):
        assert (pytest.approx(101325.0, abs=1e3) ==
                value(btx.fs.unit.outlet.pressure[0]))
        assert (pytest.approx(368.00, abs=1e-0) ==
                value(btx.fs.unit.outlet.temperature[0]))
        assert (pytest.approx(1.0, abs=1e-2) ==
                value(btx.fs.unit.outlet.flow_mol[0]))
        assert (pytest.approx(0.396, abs=1e-3) ==
                value(btx.fs.unit.control_volume.
                      properties_out[0].flow_mol_phase["Vap"]))

    @pytest.mark.ui
    @pytest.mark.unit
    def test_report(self, btx):
        btx.fs.unit.report()


# -----------------------------------------------------------------------------
@pytest.mark.iapws
@pytest.mark.skipif(not iapws95.iapws95_available(),
                    reason="IAPWS not available")
class TestIAPWS(object):
    @pytest.fixture(scope="class")
    def iapws(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = iapws95.Iapws95ParameterBlock(default={
                "phase_presentation": iapws95.PhaseType.LG})

        m.fs.unit = FeedFlash(default={"property_package": m.fs.properties,
                                       "flash_type": FlashType.isenthalpic})

        m.fs.unit.flow_mol.fix(100)
        m.fs.unit.enth_mol.fix(24000)
        m.fs.unit.pressure.fix(101325)

        return m

    @pytest.mark.unit
    def test_build(self, iapws):
        assert hasattr(iapws.fs.unit, "flow_mol")
        assert hasattr(iapws.fs.unit, "enth_mol")
        assert hasattr(iapws.fs.unit, "pressure")

        assert hasattr(iapws.fs.unit, "outlet")
        assert len(iapws.fs.unit.outlet.vars) == 3
        assert hasattr(iapws.fs.unit.outlet, "flow_mol")
        assert hasattr(iapws.fs.unit.outlet, "enth_mol")
        assert hasattr(iapws.fs.unit.outlet, "pressure")

        assert hasattr(iapws.fs.unit, "isenthalpic")

        assert number_variables(iapws) == 6
        assert number_total_constraints(iapws) == 3
        assert number_unused_variables(iapws) == 0

    @pytest.mark.component
    def test_units(self, iapws):
        assert_units_consistent(iapws)

    @pytest.mark.unit
    def test_dof(self, iapws):
        assert degrees_of_freedom(iapws) == 0

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, iapws):
        initialization_tester(iapws)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, iapws):
        results = solver.solve(iapws)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, iapws):
        assert (pytest.approx(101325.0, abs=1e3) ==
                value(iapws.fs.unit.outlet.pressure[0]))
        assert (pytest.approx(24000, abs=1e3) ==
                value(iapws.fs.unit.outlet.enth_mol[0]))
        assert (pytest.approx(100.0, abs=1e-2) ==
                value(iapws.fs.unit.outlet.flow_mol[0]))

        assert (pytest.approx(373.12, abs=1e-2) == value(
            iapws.fs.unit.control_volume.properties_out[0].temperature))
        assert (pytest.approx(0.5953, abs=1e-4) == value(
            iapws.fs.unit.control_volume.properties_out[0].phase_frac["Liq"]))

    @pytest.mark.ui
    @pytest.mark.unit
    def test_report(self, iapws):
        iapws.fs.unit.report()



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
This module contains miscalaneous utility functions for use in IDAES models.
"""

import pytest

from pyomo.environ import (Block,
                           ConcreteModel,
                           Constraint,
                           Expression,
                           Objective,
                           Set,
                           Var,
                           TransformationFactory)
from pyomo.dae import ContinuousSet, DerivativeVar
from pyomo.common.collections import ComponentSet

from idaes.core.util.model_statistics import *


# Author: Andrew Lee
@pytest.fixture()
def m():
    m = ConcreteModel()

    m.s = Set(initialize=["a", "b"])
    m.cs = ContinuousSet(bounds=(0, 1))

    m.v = Var(m.cs, initialize=1, bounds=(0, 10))
    m.dv = DerivativeVar(m.v)

    m.discretizer = TransformationFactory("dae.finite_difference")
    m.discretizer.apply_to(m,
                           nfe=10,
                           wrt=m.cs,
                           scheme="BACKWARD")

    m.e = Expression(expr=m.v[1])

    m.b1 = Block()
    m.b1.v1 = Var(initialize=1)
    m.b1.v2 = Var(m.s, initialize=1)
    m.b1.v1.fix(1)
    m.b1.c1 = Constraint(expr=1 == m.b1.v1)
    m.b1.c2 = Constraint(expr=1 <= m.b1.v1)

    m.b1.sb = Block()
    m.b1.sb.v1 = Var(initialize=1)
    m.b1.sb.v2 = Var(m.s, initialize=1)
    m.b1.sb.v1.fix(1)
    m.b1.sb.e1 = Expression(expr=m.b1.sb.v1)
    m.b1.sb.o1 = Objective(expr=1 <= m.b1.sb.v1)
    m.b1.sb.o2 = Objective(expr=1 <= m.b1.sb.v1)
    m.b1.sb.o2.deactivate()
    m.b1.sb.c1 = Constraint(expr=1 == m.b1.sb.v1)
    m.b1.sb.c2 = Constraint(expr=1 >= m.b1.sb.v1)

    m.b1.deactivate()

    m.b2 = Block(m.s)
    for i in m.s:
        m.b2[i].v1 = Var(initialize=1, bounds=(1, 10))
        m.b2[i].v2 = Var(m.s, initialize=1, bounds=(0, 1))
        m.b2[i].v1.fix(1)
        m.b2[i].e1 = Expression(expr=m.b2[i].v1)
        m.b2[i].c1 = Constraint(expr=2 == m.b2[i].v1)
        m.b2[i].c2 = Constraint(expr=2 <= m.b2[i].v1)

        if i == "a":
            m.b2[i].o1 = Objective(expr=1 <= m.b2[i].v1)
            m.b2[i].o2 = Objective(expr=1 <= m.b2[i].v1)
            m.b2[i].o2.deactivate()
            m.b2[i].c1.deactivate()
            m.b2[i].c2.deactivate()

    return m


# -------------------------------------------------------------------------
# Block methods
@pytest.mark.unit
def test_total_blocks_set(m):
    assert len(total_blocks_set(m)) == 5


@pytest.mark.unit
def test_number_total_blocks(m):
    assert number_total_blocks(m) == 5


@pytest.mark.unit
def test_activated_blocks_set(m):
    assert len(activated_blocks_set(m)) == 3


@pytest.mark.unit
def test_number_activated_blocks(m):
    assert number_activated_blocks(m) == 3


@pytest.mark.unit
def test_deactivated_blocks_set(m):
    assert len(deactivated_blocks_set(m)) == 2


@pytest.mark.unit
def test_number_deactivated_blocks(m):
    assert number_deactivated_blocks(m) == 2


# -------------------------------------------------------------------------
# Basic Constraint methods
@pytest.mark.unit
def test_total_constraints_set(m):
    assert len(total_constraints_set(m)) == 14


@pytest.mark.unit
def test_number_total_constraints(m):
    assert number_total_constraints(m) == 14


@pytest.mark.unit
def test_activated_constraints_set(m):
    assert len(activated_constraints_set(m)) == 12


@pytest.mark.unit
def test_number_activated_constraints(m):
    assert number_activated_constraints(m) == 12


@pytest.mark.unit
def test_deactivated_constraints_set(m):
    assert len(deactivated_constraints_set(m)) == 2


@pytest.mark.unit
def test_number_deactivated_constraints(m):
    assert number_deactivated_constraints(m) == 2


# -------------------------------------------------------------------------
# Equality Constraints
@pytest.mark.unit
def test_total_equalities_set(m):
    assert len(total_equalities_set(m)) == 12


@pytest.mark.unit
def test_number_total_equalities(m):
    assert number_total_equalities(m) == 12


@pytest.mark.unit
def test_activated_equalities_set(m):
    assert len(activated_equalities_set(m)) == 11


@pytest.mark.unit
def test_number_activated_equalities(m):
    assert number_activated_equalities(m) == 11


@pytest.mark.unit
def test_deactivated_equalities_set(m):
    assert len(deactivated_equalities_set(m)) == 1


@pytest.mark.unit
def test_number_deactivated_equalities(m):
    assert number_deactivated_equalities(m) == 1


# -------------------------------------------------------------------------
# Inequality Constraints
@pytest.mark.unit
def test_total_inequalities_set(m):
    assert len(total_inequalities_set(m)) == 2


@pytest.mark.unit
def test_number_total_inequalities(m):
    assert number_total_inequalities(m) == 2


@pytest.mark.unit
def test_activated_inequalities_set(m):
    assert len(activated_inequalities_set(m)) == 1


@pytest.mark.unit
def test_number_activated_inequalities(m):
    assert number_activated_inequalities(m) == 1


@pytest.mark.unit
def test_deactivated_inequalities_set(m):
    assert len(deactivated_inequalities_set(m)) == 1


@pytest.mark.unit
def test_number_deactivated_inequalities(m):
    assert number_deactivated_inequalities(m) == 1


# -------------------------------------------------------------------------
# Basic Variable Methods
# Always use ComponentSets for Vars to avoid duplication of References
# i.e. number methods should alwys use the ComponentSet, not a generator
@pytest.mark.unit
def test_variables_set(m):
    assert len(variables_set(m)) == 28


@pytest.mark.unit
def test_number_variables(m):
    assert number_variables(m) == 28


@pytest.mark.unit
def test_fixed_variables_set(m):
    assert len(fixed_variables_set(m)) == 2


@pytest.mark.unit
def test_number_fixed_variables(m):
    assert number_fixed_variables(m) == 2


@pytest.mark.unit
def test_unfixed_variables_set(m):
    assert len(unfixed_variables_set(m)) == 26


@pytest.mark.unit
def test_number_unfixed_variables(m):
    assert number_unfixed_variables(m) == 26


@pytest.mark.unit
def test_variables_near_bounds_set(m):
    tset = variables_near_bounds_set(m)
    assert len(tset) == 6
    for i in tset:
        assert i in ComponentSet(
            [m.b2["a"].v1, m.b2["b"].v1, m.b2["a"].v2["a"],
             m.b2["a"].v2["b"], m.b2["b"].v2["a"], m.b2["b"].v2["b"]])

    m.b2["a"].v1.value = 1.001
    tset = variables_near_bounds_set(m)
    assert len(tset) == 5
    for i in tset:
        assert i in ComponentSet(
            [m.b2["b"].v1, m.b2["a"].v2["a"],
             m.b2["a"].v2["b"], m.b2["b"].v2["a"], m.b2["b"].v2["b"]])

    tset = variables_near_bounds_set(m, tol=1e-3)
    assert len(tset) == 6
    for i in tset:
        assert i in ComponentSet(
            [m.b2["a"].v1, m.b2["b"].v1, m.b2["a"].v2["a"],
             m.b2["a"].v2["b"], m.b2["b"].v2["a"], m.b2["b"].v2["b"]])

    m.b2["a"].v1.setlb(None)
    tset = variables_near_bounds_set(m)
    assert len(tset) == 5
    for i in tset:
        assert i in ComponentSet(
            [m.b2["b"].v1, m.b2["a"].v2["a"],
             m.b2["a"].v2["b"], m.b2["b"].v2["a"], m.b2["b"].v2["b"]])

    m.b2["a"].v2["a"].setub(None)
    tset = variables_near_bounds_set(m)
    assert len(tset) == 4
    for i in tset:
        assert i in ComponentSet(
            [m.b2["b"].v1, m.b2["a"].v2["b"],
             m.b2["b"].v2["a"], m.b2["b"].v2["b"]])

    m.b2["a"].v2["b"].value = None
    tset = variables_near_bounds_set(m)
    assert len(tset) == 3
    for i in tset:
        assert i in ComponentSet(
            [m.b2["b"].v1, m.b2["b"].v2["a"], m.b2["b"].v2["b"]])


@pytest.mark.unit
def test_number_variables_near_bounds(m):
    assert number_variables_near_bounds(m) == 6


# -------------------------------------------------------------------------
# Variables in Constraints
@pytest.mark.unit
def test_variables_in_activated_constraints_set(m):
    assert len(variables_in_activated_constraints_set(m)) == 22


@pytest.mark.unit
def test_number_variables_in_activated_constraints(m):
    assert number_variables_in_activated_constraints(m) == 22


@pytest.mark.unit
def test_variables_in_activated_equalities_set(m):
    assert len(variables_in_activated_equalities_set(m)) == 22


@pytest.mark.unit
def test_number_variables_in_activated_equalities(m):
    assert number_variables_in_activated_equalities(m) == 22


@pytest.mark.unit
def test_variables_in_activated_inequalities_set(m):
    assert len(variables_in_activated_inequalities_set(m)) == 1


@pytest.mark.unit
def test_number_variables_in_activated_inequalities(m):
    assert number_variables_in_activated_inequalities(m) == 1


@pytest.mark.unit
def test_variables_only_in_inequalities(m):
    m.v3 = Var(initialize=1)
    m.c3 = Constraint(expr=m.v3 >= 1)
    assert len(variables_only_in_inequalities(m)) == 1


@pytest.mark.unit
def test_number_variables_only_in_inequalities(m):
    m.v3 = Var(initialize=1)
    m.c3 = Constraint(expr=m.v3 >= 1)
    assert number_variables_only_in_inequalities(m) == 1


# -------------------------------------------------------------------------
# Fixed Variables in Constraints
@pytest.mark.unit
def test_fixed_variables_in_activated_equalities_set(m):
    assert len(fixed_variables_in_activated_equalities_set(m)) == 1


@pytest.mark.unit
def test_number_fixed_variables_in_activated_equalities(m):
    assert number_fixed_variables_in_activated_equalities(m) == 1


@pytest.mark.unit
def test_fixed_variables_only_in_inequalities(m):
    m.v3 = Var(initialize=1)
    m.v3.fix(1)
    m.c3 = Constraint(expr=m.v3 >= 1)
    assert len(fixed_variables_only_in_inequalities(m)) == 1


@pytest.mark.unit
def test_number_fixed_variables_only_in_inequalities(m):
    m.v3 = Var(initialize=1)
    m.v3.fix(1)
    m.c3 = Constraint(expr=m.v3 >= 1)
    assert number_fixed_variables_only_in_inequalities(m) == 1


# -------------------------------------------------------------------------
# Unused and un-Transformed Variables
@pytest.mark.unit
def test_unused_variables_set(m):
    assert len(unused_variables_set(m)) == 6


@pytest.mark.unit
def test_number_unused_variables(m):
    assert number_unused_variables(m) == 6


@pytest.mark.unit
def test_fixed_unused_variables_set(m):
    assert len(fixed_unused_variables_set(m)) == 1


@pytest.mark.unit
def test_number_fixed_unused_variables(m):
    assert number_fixed_unused_variables(m) == 1


@pytest.mark.unit
def test_derivative_variables_set():
    m = ConcreteModel()

    m.cs = ContinuousSet(bounds=(0, 1))

    m.v = Var(m.cs, initialize=1)
    m.dv = DerivativeVar(m.v)

    assert len(derivative_variables_set(m)) == 2

    m.discretizer = TransformationFactory("dae.finite_difference")
    m.discretizer.apply_to(m,
                           nfe=10,
                           wrt=m.cs,
                           scheme="BACKWARD")

    assert len(derivative_variables_set(m)) == 0


@pytest.mark.unit
def test_number_derivative_variables():
    m = ConcreteModel()

    m.cs = ContinuousSet(bounds=(0, 1))

    m.v = Var(m.cs, initialize=1)
    m.dv = DerivativeVar(m.v)

    assert number_derivative_variables(m) == 2

    m.discretizer = TransformationFactory("dae.finite_difference")
    m.discretizer.apply_to(m,
                           nfe=10,
                           wrt=m.cs,
                           scheme="BACKWARD")

    assert number_derivative_variables(m) == 0


# -------------------------------------------------------------------------
# Objective methods
@pytest.mark.unit
def test_total_objectives_set(m):
    assert len(total_objectives_set(m)) == 2


@pytest.mark.unit
def test_number_total_objectives(m):
    assert number_total_objectives(m) == 2


@pytest.mark.unit
def test_activated_objectives_set(m):
    assert len(activated_objectives_set(m)) == 1


@pytest.mark.unit
def test_number_activated_objectives(m):
    assert number_activated_objectives(m) == 1


@pytest.mark.unit
def test_deactivated_objectives_set(m):
    assert len(deactivated_objectives_set(m)) == 1


@pytest.mark.unit
def test_number_deactivated_objectives(m):
    assert number_deactivated_objectives(m) == 1


# -------------------------------------------------------------------------
# Expression methods
@pytest.mark.unit
def test_expressions_set(m):
    assert len(expressions_set(m)) == 3


@pytest.mark.unit
def test_number_expressions(m):
    assert number_expressions(m) == 3


# -------------------------------------------------------------------------
# Other model statistics
@pytest.mark.unit
def test_degrees_of_freedom(m):
    assert degrees_of_freedom(m) == 10


@pytest.mark.unit
def test_large_residuals_set(m):
    # Initialize derivative var values so no errors occur
    for v in m.dv.keys():
        m.dv[v] = 0
    assert len(large_residuals_set(m)) == 2


@pytest.mark.unit
def test_number_large_residuals(m):
    # Initialize derivative var values so no errors occur
    for v in m.dv.keys():
        m.dv[v] = 0
    assert number_large_residuals(m) == 2


@pytest.mark.unit
def test_active_variables_in_deactivated_blocks_set(m):
    assert len(active_variables_in_deactivated_blocks_set(m)) == 0

    m.c = Constraint(expr=m.b1.v1 >= 2)

    assert len(active_variables_in_deactivated_blocks_set(m)) == 1


@pytest.mark.unit
def test_number_active_variables_in_deactivated_blocks(m):
    assert number_active_variables_in_deactivated_blocks(m) == 0

    m.c = Constraint(expr=m.b1.v1 >= 2)

    assert number_active_variables_in_deactivated_blocks(m) == 1


# -------------------------------------------------------------------------
# Reporting methods
@pytest.mark.unit
def test_report_statistics(m):
    report_statistics(m)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Install IDAES example files locally.

By default, this will download the examples from "examples-pse/src" on Github,
with a release matching the version of the currently installed idaes package,
into a sub-directory of the current directory named "examples". It will
also, unless directed otherwise, install all Python modules from the downloaded
directory into a package called "idaes_examples".

Options let the user choose a different version, directory, and
whether to actually download or install.
"""
# Pyomo utility for delayed import
from pyomo.common.dependencies import attempt_import

# stdlib
from collections import namedtuple
from datetime import datetime
from io import StringIO
import logging
from operator import attrgetter
import os
import re
from pathlib import Path
import shutil
import sys
from typing import List
from uuid import uuid4
from zipfile import ZipFile
import json


# third-party
import click
# third-party slow
nb_exporters= attempt_import("nbconvert.exporters")[0]
nb_writers = attempt_import("nbconvert.writers")[0]
traitlets_config = attempt_import("traitlets.config")[0]
nbformat = attempt_import("nbformat")[0]
requests = attempt_import("requests")[0]

# package
from idaes.commands import cb
from idaes.commands.base import how_to_report_an_error
from idaes.ver import package_version as V

__author__ = "Dan Gunter"


_log = logging.getLogger("idaes.commands.examples")

# Constants

GITHUB = "https://github.com"
GITHUB_API = "https://api.github.com"
REPO_ORG = "idaes"
REPO_NAME = "examples-pse"
REPO_DIR = "src"
PKG_VERSION = f"{V.major}.{V.minor}.{V.micro}"
INSTALL_PKG = "idaes_examples"
JUPYTER_NB_VERSION = 4  # for parsing

REMOVE_CELL_TAG = "remove_cell"  # for stripping Jupyter notebook cells
SOLUTION_CELL_TAG = "solution"  # ditto
EXERCISE_CELL_TAG = "exercise"  # ditto

STRIPPED_NOTEBOOK_SUFFIX = "_s"

# Global vars

g_tempdir, g_egg = None, None

# Exceptions


class DownloadError(Exception):
    """Used for errors downloading the release files.
    """

    pass


class CopyError(Exception):
    """Used for errors copying files.
    """
    pass


class InstallError(Exception):
    """Used for errors installing the source as a Python package.
    """

    pass


class GithubError(Exception):
    pass


Release = namedtuple("Release", ["date", "tag", "info"])


@cb.command(
    name="get-examples", help="Fetch example scripts and Jupyter Notebooks."
)
@click.option(
    "--dir", "-d", "directory", help="installation target directory", default="examples",
    type=str,
)
@click.option(
    "--local", "local_dir",
    help="For developers: instead of downloading, copy from an "
         "idaes-examples repository on local disk"
)
@click.option(
    "--no-install", "-I", "no_install", help="Do *not* install examples into 'idaes_examples' package",
    is_flag=True
)
@click.option(
    "--list",
    "-l",
    "list_releases",
    help="List all available released versions, and stop",
    is_flag=True,
)
@click.option(
    "--no-download",
    "-N",
    "no_download",
    help="Do not download anything",
    is_flag=True
)
@click.option(
    "--unstable",
    "-U",
    help="Allow and list unstable/pre-release versions",
    is_flag=True
)
@click.option(
    "--version",
    "-V",
    help=f"Version of examples to download",
    default=None,
    show_default=False,
)
def get_examples(directory, no_install, list_releases, no_download, version,
                 unstable, local_dir):
    """Get the examples from Github and put them in a local directory.
    """
    # list-releases mode
    if list_releases:
        try:
            releases = get_releases(unstable)
        except GithubError as err:
            click.echo(f"Error getting data from Github: {err}")
            sys.exit(-1)
        print_releases(releases, unstable)
        sys.exit(0)
    # otherwise..
    target_dir = Path(directory)
    # do nothing?
    if no_download and no_install:
        click.echo("Download and installation disabled. Done.")
        sys.exit(0)
    # download?
    if no_download:  # no download
        _log.info("skipping download step (use existing examples)")
        click.echo("Skip download")
    else:  # download
        if target_dir.exists():
            click.echo(f"Target directory '{target_dir}' already exists. Please "
                       f"remove it, or choose a different directory.")
            sys.exit(-1)
        if local_dir is not None:
            click.echo(f"Copying from local directory: {local_dir}")
            local_path = Path(local_dir)
            if not local_path.exists() or not local_path.is_dir():
                click.echo(f"Cannot copy from local directory '{local_dir}': "
                           f"directory does not exist, or not a directory")
                sys.exit(-1)
            try:
                copy_contents(target_dir, local_path)
            except CopyError as err:
                click.echo(f"Failed to copy from '{local_dir}' to '{target_dir}': "
                           f"{err}")
                sys.exit(-1)
            ex_version = version if version else PKG_VERSION
        else:
            try:
                releases = get_releases(unstable)
            except GithubError as err:
                click.echo(f"Error getting data from Github: {err}")
                sys.exit(-1)
            # check that unstable flag is given before downloading unstable version
            if version is not None:
                stable_ver = re.match(r".*-\w+$", version) is None
                if not stable_ver and not unstable:
                    click.echo(f"Cannot download unstable version {version} unless you add "
                               f"the -U/--unstable flag")
                    sys.exit(-1)
            # set version
            if version is None:
                ex_version = get_examples_version(PKG_VERSION)
            else:
                ex_version = version
            # give an error if selected version does not exist
            if ex_version not in [r.tag for r in releases]:
                if version is None:
                    click.echo(f"Internal Error: Could not find an examples release\n"
                               f"matching IDAES version {PKG_VERSION}\n"
                               f"You can manually pick  version with '-V/--version'\n"
                               f"or install from a local directory with '--local'.\n"
                               f"Use '-l/--list-releases' to see all versions.\n"
                               f"{how_to_report_an_error()}\n")
                else:
                    click.echo(f"Could not find an examples release matching IDAES version "
                               f"{version}.\n Use -l/--list-releases to see all "
                               f"available versions.")
                sys.exit(-1)
            click.echo("Downloading...")
            try:
                download(target_dir, ex_version)
            except DownloadError as err:
                _log.warning(f"abort due to failed download: {err}")
                clean_up_temporary_files()
                sys.exit(-1)
            full_dir = os.path.realpath(target_dir)
            _log.info(f"downloaded examples to: '{full_dir}'")
    # install
    if no_install:
        _log.info("skipping installation step")
        click.echo("Skip install")
    else:
        if not target_dir.exists():
            if no_download:
                click.echo(f"Target directory '{target_dir}' does not exist")
                sys.exit(-1)
            else:
                click.echo(f"Internal error: After download, directory '{target_dir}'\n"
                           f"does not exist.\n"
                           f"{how_to_report_an_error()}")
                sys.exit(-1)
        click.echo("Installing...")
        try:
            install_src(ex_version, target_dir)
        except InstallError as err:
            click.echo(f"Install error: {err}")
            clean_up_temporary_files()
            sys.exit(-1)
        _log.info(f"Installed examples as package {INSTALL_PKG}")

    click.echo("Cleaning up...")
    # strip notebooks
    _log.info("Stripping test/solution cells from notebooks")
    strip_special_cells(target_dir)
    # temporary files
    _log.info("Removing temporary files")
    clean_up_temporary_files()

    # Done
    print_summary( ex_version, target_dir, not no_install)


def print_summary(version, dirname, installed):
    sep1, sep2 = "-" * 40, "=" * 40
    print(f"{sep1}\nIDAES Examples {version}\n{sep2}")
    print(f"Path   : {dirname}")
    if installed:
        print(f"Package: {INSTALL_PKG}")
    else:
        print(f"Package: not installed")
    print(sep2)


def get_examples_version(idaes_version: str):
    """Given the specified 'idaes-pse' repository release version,
    identify the matching 'examples-pse' repository release version.

    Args:
        idaes_version: IDAES version, e.g. "1.5.0" or "1.5.0.dev0+e1bbb[...]"

    Returns:
        Examples version, or if there is no match, return None.
    """
    # Fetch the idaes:examples version mapping from Github
    compat_file = 'idaes-compatibility.json'
    url = f"{GITHUB_API}/repos/{REPO_ORG}/{REPO_NAME}/contents/{compat_file}"
    headers = {'Accept': 'application/vnd.github.v3.raw'}
    _log.debug(f'About to call requests.get({url}, {headers})')
    res = requests.get(url, headers=headers)
    if not res.ok:
        _log.debug(f'Problem getting mapping file: {res.json()}')
        raise DownloadError(res.json())

    try:
        compat_mapping = json.loads(res.text)['mapping']
    except KeyError:
        # return the latest version instead
        _log.warning('Ill-formed compatibility mapping file for examples repository:')
        _log.debug(f'compat_mapping: {res.text}')
        _log.info('Defaulting to latest released version of examples.')
        return None

    idaes_version_num = idaes_version
    version_numbers = idaes_version.split('.')
    if len(version_numbers) > 3:
        idaes_version_num = '.'.join(version_numbers[:3])
        click.echo(f"Warning: non-release version of IDAES detected. "
                   f"Using IDAES {idaes_version_num} as reference; "
                   f"examples version compatibility is not guaranteed.")

    try:
        examples_version = compat_mapping[idaes_version_num]
    except KeyError:
        # return the latest version instead, as above
        _log.warning('IDAES version not found in compatibility mapping file. \
                Defaulting to latest released version of examples.')
        return None

    _log.debug(f'get_examples_version({idaes_version}: {examples_version}')

    return examples_version


def download(target_dir: Path, version: str):
    """Download `version` into `target_dir`.
    Raises:
        DownloadError
    """
    # check target directory
    if target_dir.exists():
        click.echo(f"Directory '{target_dir}' exists. Please move or delete and "
                   f"try this command again")
        raise DownloadError("directory exists")
    # download
    try:
        download_contents(target_dir, version)
    except DownloadError as err:
        click.echo(f"Download failed: {err}")
        raise


def is_illegal_dir(d: Path):
    """Refuse to remove directories for some situations, for safety.
    """
    if (d / ".git").exists():
        return ".git file found"
    if d.absolute() == Path.home().absolute():
        return "cannot replace home directory"
    if d.absolute == Path("/").absolute():
        return "cannot replace root directory"
    return None


def download_contents(target_dir, version):
    """Download the given version from the Github releases and make
    its `REPO_DIR` subdirectory be the `target_dir`.

    Raises:
        DownloadError: if the GET on the release URL returns non-200 status
    """
    global g_tempdir
    url = archive_file_url(version)
    _log.info(f"get examples from: {url}")
    # stream out to a big .zip file
    req = requests.get(url, stream=True)
    if req.status_code != 200:
        if req.status_code in (400, 404):
            raise DownloadError(f"file not found")
        raise DownloadError(f"status={req.status_code}")
    # note: mkdtemp() creates a directory that seems un-removable on Windows.
    # So, instead, just create the directory yourself in the current directory
    random_name = str(uuid4())
    try:
        os.mkdir(random_name)
    except Exception as err:
        _log.fatal(f"making directory '{random_name}': {err}")
        click.echo("Cannot make temporary directory in current directory. Abort.")
        sys.exit(-1)
    g_tempdir = tempdir = Path(random_name)
    _log.debug(f"created temporary directory '{tempdir.name}'")
    tempfile = tempdir / "examples.zip"
    with tempfile.open("wb") as f:
        for chunk in req.iter_content(chunk_size=65536):
            f.write(chunk)
    _log.info(f"downloaded zipfile to {tempfile}")
    # open as a zip file, and extract all files into the temporary directory
    _log.debug(f"open zip file: {tempfile}")
    zipf = ZipFile(str(tempfile))
    zipf.extractall(path=tempdir.name)
    # move the REPO_DIR subdirectory into the target dir
    subdir = Path(tempdir.name) / f"{REPO_NAME}-{version}" / REPO_DIR
    _log.debug(f"move {subdir} -> {target_dir}")
    shutil.move(str(subdir), str(target_dir))
    zipf.close()


def copy_contents(target_dir, repo_root):
    subdir = repo_root / REPO_DIR
    if not subdir.is_dir():
        raise CopyError(f"Could not copy from '{subdir}': not a directory")
    _log.info(f"copy.local.start from={subdir} to={target_dir}")
    try:
        shutil.copytree(subdir, target_dir)
    except shutil.Error as err:
        raise CopyError(err)
    except FileNotFoundError:
        raise CopyError(f"Could not find file '{subdir}'")
    except Exception as err:
        raise CopyError(f"Unknown problem copying: {err}")
    _log.info(f"copy.local.end from={subdir} to={target_dir}")


def clean_up_temporary_files():
    # temporary directory created for unzipping and renaming
    if g_tempdir:
        d = g_tempdir
        _log.debug(f"remove temporary directory.start name='{d}'")
        try:
            shutil.rmtree(d)
        except Exception as err:  # WTF
            _log.warning(f"remove temporary directory.error name='{d}' msg='{err}'")
        else:
            _log.debug(f"removed temporary directory.end name='{d}'")
    # egg file created by setuptools
    if g_egg and g_egg.exists() and g_egg.is_dir():
        _log.debug(f"remove setuptools egg path='{g_egg}'")
        try:
            shutil.rmtree(g_egg.name)
        except Exception as err:
            _log.warning(f"remove temporary directory.error name='{g_egg}' msg='{err}'")
    # dist directory created by setuptools
    d = Path("dist")
    if d.exists() and d.is_dir():
        for f in d.glob("idaes_examples-*.egg"):
            try:
                f.unlink()
            except Exception as err:
                _log.warning(f"could not remove distribution file {f}: {err}")
        # remove directory, if now empty
        num_files = len(list(d.glob("*")))
        if num_files == 0:
            _log.info(f"removing dist directory '{d.absolute()}'")
            try:
                d.rmdir()
            except Exception as err:
                _log.warning(f"could not remove distribution directory {d}: {err}")


def archive_file_url(version, org=REPO_ORG, repo=REPO_NAME):
    """Build & return URL for a given release version.
    """
    return f"{GITHUB}/{org}/{repo}/archive/{version}.zip"


def get_releases(unstable) -> List[Release]:
    """Returns a list of releases.

    The list is sorted in ascending order by date.
    """
    releases = []
    url = f"{GITHUB_API}/repos/{REPO_ORG}/{REPO_NAME}/releases"
    resp = requests.get(url)
    data, headers = resp.json(), resp.headers
    check_github_response(data, headers)
    for rel in data:
        if not unstable and rel["prerelease"]:
            continue
        releases.append(Release(rel["published_at"], rel["tag_name"], rel["name"]))
    releases.sort(key=attrgetter("date"))  # sort by publication date
    return releases


def check_github_response(data, headers):
    """Check whether GitHub gave an error message. If so, raise a GithubError
    with a hopefully informative and useful message.
    """
    if isinstance(data, list):
        return  # lists are assumed to be the releases
    if isinstance(data, dict) and "message" in data:
        if "rate limit exceeded" in data["message"]:
            reset_ts = int(headers["X-RateLimit-Reset"])
            now = datetime.now()
            now_ts, tzinfo = now.timestamp(), now.astimezone().tzinfo
            wait_min = int((reset_ts - now_ts) // 60) + 1
            reset_dt = datetime.fromtimestamp(reset_ts)
            datestr = reset_dt.astimezone(tzinfo).isoformat()
            raise GithubError(f"API rate limit exceeded.\n"
                              f"You will need to wait {wait_min} minutes,"
                              f" until {datestr}, to try again from "
                              f"this computer.")
        else:
            raise GithubError(f"Error connecting to Github: {data['message']}")
    else:
        raise GithubError(f"Invalid result from Github: data={data} headers={headers}")


def print_releases(releases: List[Release], unstable):
    """Print the releases, as returned by `get_releases()`, as a table
    to standard output.
    """
    if len(releases) == 0:
        if unstable:
            print("No releases found")
        else:
            print("No stable releases found. Add -U/--unstable to also look "
                  "for pre-releases.")
        return
    # determine column widths
    widths = [4, 7, 7]  # widths of column titles: date,version,details
    widths[0] = len(releases[0].date)  # dates are all the same
    # tags and names can have different widths
    for rel in releases:
        for i in range(1, 3):
            widths[i] = max(widths[i], len(rel[i]))
    # make row format
    pad = "  "
    fmt = f"{{date:{widths[0]}s}}{pad}{{tag:{widths[1]}s}}{pad}{{name:{widths[2]}s}}"
    # print header
    print("")
    print(fmt.format(date="Date", tag="Version", name="Details"))
    print(fmt.format(date="-" * widths[0], tag="-" * widths[1], name="-" * widths[2]))
    # print rows
    for rel in releases:
        print(fmt.format(date=rel.date, tag=rel.tag, name=rel.info))
    # print footer
    print("")


def install_src(version, target_dir):
    """Install the 'src' subdirectory as a package, given by `INSTALL_PKG`,
    by renaming the directory, adding '__init__.py' files,
    and then running `setuptools.setup()` on the directory tree.
    When done, name the directory back to 'src', and remove '__init__.py' files.
    Then clean up whatever cruft is left behind..
    """
    from setuptools import setup, find_packages # import here due to slowness
    global g_egg
    orig_dir = Path(os.curdir).absolute()
    target_dir = Path(target_dir.absolute())
    root_dir = target_dir.parent
    examples_dir = root_dir.absolute() / INSTALL_PKG
    if examples_dir.exists():
        raise InstallError(f"package directory {examples_dir} already exists")
    _log.info(f"install into {INSTALL_PKG} package")
    # set the args to make it look like the 'install' command has been invoked
    saved_args = sys.argv[:]
    sys.argv = ["setup.py", "install"]
    # add some empty __init__.py files
    _log.debug("add temporary __init__.py files")
    pydirs = find_python_directories(target_dir)
    pydirs.append(target_dir)  # include top-level dir
    try:
        for d in pydirs:
            init_py = d / "__init__.py"
            init_py.open("w")
    except IOError as err:
        raise InstallError(f"error writing temporary __init__.py files: {err}")
    # temporarily rename target directory to the package name
    _log.info(f"rename {target_dir} -> {examples_dir}")
    shutil.move(target_dir, examples_dir)
    # if there is a 'build' directory, move it aside
    build_dir = root_dir / 'build'
    if build_dir.exists():
        from uuid import uuid1
        random_letters = str(uuid1())
        moved_build_dir = f"{build_dir}.{random_letters}"
        _log.debug(f"move existing build dir to {moved_build_dir}")
        shutil.move(str(build_dir), moved_build_dir)
    else:
        _log.debug("no existing build directory (nothing to do)")
        moved_build_dir = None
    # run setuptools' setup command (in root directory)
    _log.info(f"run setup command in directory {root_dir}")
    os.chdir(root_dir)
    packages = [d for d in find_packages() if d.startswith(INSTALL_PKG)]
    _log.debug(f"install packages: {packages}")
    # before running, grab stdout
    orig_stdout = sys.stdout
    sys.stdout = setup_out = StringIO()
    # run setup
    setup(
        name=INSTALL_PKG,
        version=version,
        # description='IDAES examples',
        packages=packages,
        python_requires=">=3.5,  <4",
        zip_safe=False
    )
    # restore stdout
    sys.stdout = orig_stdout
    # print/log output
    output_str = setup_out.getvalue()
    if _log.isEnabledFor(logging.DEBUG):
        for line in output_str.split("\n"):
            _log.debug(f"(setup) {line}")
    # name the target directory back to original
    _log.info(f"rename '{examples_dir}' to  '{target_dir}'")
    shutil.move(examples_dir, target_dir)
    # remove the empty __init__.py files
    _log.info("remove temporary __init__.py files")
    for d in pydirs:
        init_py = d / "__init__.py"
        _log.debug(f"remove '{init_py}")
        init_py.unlink()
    # remove build dir, and restore any moved build dir
    _log.info(f"remove build directory '{build_dir}'")
    try:
        shutil.rmtree(build_dir)
    except Exception as err:
        _log.warning(f"failed to remove build directory {build_dir}: {err}")
    if moved_build_dir is not None:
        _log.info(f"restore build dir '{build_dir}' from '{moved_build_dir}'")
        shutil.move(moved_build_dir, build_dir)
    # restore previous args
    sys.argv = saved_args
    # change back to previous directory
    os.chdir(orig_dir)
    # save name of egg file, for later cleanup
    f = root_dir / (INSTALL_PKG + ".egg-info")
    if f.exists():
        g_egg = f
    else:
        _log.warning(f"egg-info file not found path='{f}'")


def find_python_directories(target_dir: Path) -> List[Path]:
    """Find all directories from target_dir, on down, that contain a
    Python module or sub-package.
    """
    # get directories that contain python files -> pydirs
    pydirs = set((x.parent for x in target_dir.rglob("*.py")))
    # get all directories in the tree leading to the 'pydirs'
    alldirs = set()
    for d in pydirs:
        while d != target_dir:
            alldirs.add(d)
            d = d.parent
    return list(alldirs)


def find_notebook_files(target_dir: Path) -> List[Path]:
    """Find all files ending in ".ipynb", at or below target_dir.
    """
    return list(target_dir.rglob("*.ipynb"))


def strip_special_cells(target_dir: Path):
    """Strip 'special' cells from notebooks below `target_dir`.
    See `strip_tags()` for how that is done.
    """
    nb_files = find_notebook_files(target_dir)
    for nb_file in nb_files:
        if strip_tags(nb_file):
            _log.info(f"removing original file '{nb_file}'")
            nb_file.unlink()


def strip_tags(nb: Path) -> bool:
    """Strip tags from notebook, if there are any.

    Behavior depends on filename (case-insensitive).

    - if the notebook file name ends with "_solution_testing", create two files:

       1. chop off "_testing" from name, remove test and exercise cells
       2. chop off "_solution_testing" from name & add "_exercise", remove solution and test cells

    - if the file ends with "_testing" only, create one file:

        1. chop off "_testing" from name, remove test cells

    - otherwise: do nothing

    Returns:
        Was anything stripped (and new files created)?

    Raises:
        IOError: If the desired target for the stripped notebook already exists
    """
    _log.info(f"stripping tags from notebook {nb}")
    nb_name = nb.stem

    # Figure out which tags to strip:
    # - check for case (1), solution_testing.ipynb
    if nb_name.lower().endswith("_solution_testing"):
        pre = nb_name[:-17]
        names = [f"{pre}_solution.ipynb", f"{pre}_exercise.ipynb"]
        tags_to_strip = [(REMOVE_CELL_TAG, EXERCISE_CELL_TAG), (REMOVE_CELL_TAG, SOLUTION_CELL_TAG)]
    # - check for case (2), _testing.ipynb
    elif nb_name.lower().endswith("_testing"):
        pre = nb_name[:-8]
        names = [f"{pre}.ipynb"]
        tags_to_strip = [(REMOVE_CELL_TAG,)]
    # - if neither, we are done here
    else:
        _log.debug(f"notebook '{nb}' does not need to have tags stripped")
        return False

    # Create all desired tag-stripped copies
    for name, remove_tags in zip(names, tags_to_strip):
        target_nb = nb.parent / name
        _log.info(f"creating new notebook '{target_nb}'")
        # Don't overwrite an existing file
        if target_nb.exists():
            _log.warning(f"cannot create new notebook '{target_nb}': file exists")
            continue
        # Set up configuration for removing specially tagged cells
        conf = traitlets_config.Config()
        conf.TagRemovePreprocessor.remove_cell_tags = remove_tags
        conf.NotebookExporter.preprocessors = [
            # this requires the full module path
            "nbconvert.preprocessors.TagRemovePreprocessor"
        ]
        # Convert from Notebook format to Notebook format, stripping tags
        (body, resources) = nb_exporters.NotebookExporter(
            config=conf).from_filename(str(nb))
        # Set up output destination
        wrt = nb_writers.FilesWriter()
        wrt.build_directory = str(target_nb.parent)
        # Write stripped notebook to output file
        wrt.write(body, resources, notebook_name=target_nb.stem)

    return True

def has_tagged_cells(nb: Path):
    """Quickly check whether this notebook has any cells with the "special" tag.

    Returns:
        True = yes, it does; False = no specially tagged cells
    Raises:
        NotebookFormatError, if notebook at 'entry' is not parseable
    """
    # parse the notebook (assuming this is fast; otherwise should cache it)
    try:
        nb = nbformat.read(str(nb), as_version=JUPYTER_NB_VERSION)
    except nbformat.reader.NotJSONError:
        raise ValueError(f"Notebook '{nb}' is not valid JSON")
    # look for tagged cells; return immediately if one is found
    for i, c in enumerate(nb.cells):
        if "tags" in c.metadata and REMOVE_CELL_TAG in c.metadata.tags:
            _log.debug(f"Found {REMOVE_CELL_TAG} tag in cell {i}")
            return True  # can stop now, one is enough
    # no tagged cells
    return False



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Unit operation model for a steam heater applicable to platen superheater
and roof superheater, model main equations:

* Heat is given by fire-side boiler model
* Calculate pressure change due to friction
* Calculate slag layer wall temperature
* Consider a layer of metal and a layer of slag

Created on Aug 27, 2020 by Boiler Team (J. Ma, M. Zamarripa)
"""
import pytest
# Import Pyomo libraries
import pyomo.environ as pyo

# Import IDAES core
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom

# Import Unit Model Modules
from idaes.generic_models.properties import iapws95

from idaes.core.util.testing import initialization_tester
from idaes.core.util import get_solver
from idaes.power_generation.unit_models.steamheater import SteamHeater

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()

# -----------------------------------------------------------------------------


@pytest.fixture(scope="module")
def build_unit():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit = SteamHeater(default={
        "dynamic": False,
        "property_package": m.fs.properties,
        "has_holdup": True,
        "has_heat_transfer": True,
        "has_pressure_change": True,
        "single_side_only": True}
        )
    return m


@pytest.mark.unit
def test_basic_build(build_unit):
    """Make a steam heater model and make sure it doesn't throw exception"""
    m = build_unit
    assert degrees_of_freedom(m) == 12
    # Check unit config arguments
    assert len(m.fs.unit.config) == 10
    assert m.fs.unit.config.has_heat_transfer
    assert isinstance(m.fs.unit.heat_duty, pyo.Var)
    assert isinstance(m.fs.unit.deltaP, pyo.Var)
    assert m.fs.unit.config.property_package is m.fs.properties


@pytest.mark.skipif(not iapws95.iapws95_available(),
                    reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_initialize_unit(build_unit):
    m = build_unit
    # platen superheater
    m.fs.unit.diameter_in.fix(0.035)
    m.fs.unit.tube_thickness.fix(0.005)
    m.fs.unit.fin_thickness.fix(0.0035)
    m.fs.unit.slag_thickness[:].fix(0.001)
    m.fs.unit.fin_length.fix(0.00966)
    m.fs.unit.tube_length.fix(40.0)
    m.fs.unit.number_tubes.fix(200.0)
    m.fs.unit.therm_cond_slag.fix(1.3)
    m.fs.unit.control_volume.scaling_factor_holdup_energy = 1e-8
    m.fs.unit.heat_fireside[:].fix(5.5e7)  # initial guess
    hpl = iapws95.htpx(798.15*pyo.units.K, 24790249.01*pyo.units.Pa)
    m.fs.unit.inlet[:].flow_mol.fix(24194.177)
    m.fs.unit.inlet[:].enth_mol.fix(hpl)
    m.fs.unit.inlet[:].pressure.fix(24790249.01)

    state_args = {'flow_mol': 24194.177,
                  'pressure': 24790249.01,
                  'enth_mol': hpl}

    initialization_tester(build_unit, dof=0, state_args=state_args)


@pytest.mark.skipif(not iapws95.iapws95_available(),
                    reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_solve_unit(build_unit):
    m = build_unit

    m.fs.unit.inlet.enth_mol.fix()
    m.fs.unit.inlet.flow_mol.fix()
    m.fs.unit.inlet.pressure.fix()
    assert degrees_of_freedom(m) == 0

    results = solver.solve(m, tee=True)
    # Check for optimal solution
    assert results.solver.termination_condition == \
        pyo.TerminationCondition.optimal
    assert results.solver.status == pyo.SolverStatus.ok

    # check material balance
    assert (pytest.approx(pyo.value(m.fs.unit.control_volume.properties_in[0].
                                    flow_mol - m.fs.unit.control_volume.
                                    properties_out[0].flow_mol),
                          abs=1e-3) == 0)

    # pressure drop
    assert (pytest.approx(-282158.4020, abs=1e-3) == pyo.value(m.fs.unit.
                                                                deltaP[0]))
    # check energy balance
    assert (pytest.approx(
        pyo.value(m.fs.unit.control_volume.properties_in[0].enth_mol
                  - m.fs.unit.control_volume.properties_out[0].enth_mol),
        abs=1e-3) == -2273.2742)

    assert (pytest.approx(835.7582, abs=1e-3) ==
            pyo.value(m.fs.unit.control_volume.properties_out[0].temperature))



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Base class for EoS modules.

Raises NotImplementedErrors for all expected methods in case developer misses
some. EoS developers should overload all these methods.
"""
from pyomo.environ import units as pyunits
from idaes.core.util.constants import Constants as const
from idaes.generic_models.properties.core.generic.utility import (
    get_method, get_component_object as cobj)
from idaes.core.util.exceptions import (
    PropertyNotSupportedError, ConfigurationError)


class EoSBase():

    @staticmethod
    def gas_constant(b):
        # Utility method to convert gas constant to base units
        base_units = b.params.get_metadata().default_units

        r_units = (base_units["mass"] *
                   base_units["length"]**2 *
                   base_units["temperature"]**-1 *
                   base_units["amount"]**-1 *
                   base_units["time"]**-2)

        return pyunits.convert(const.gas_constant, to_units=r_units)

    @staticmethod
    def common(b, pobj):
        raise NotImplementedError(_msg(b, "common"))

    @staticmethod
    def calculate_scaling_factors(b, pobj):
        raise NotImplementedError(_msg(b, "calculate_scaling_factors"))

    @staticmethod
    def build_parameters(b):
        raise NotImplementedError(_msg(b, "build_parameters"))

    @staticmethod
    def cp_mol_phase(b, p):
        raise NotImplementedError(_msg(b, "cp_mol_phase"))

    @staticmethod
    def cp_mol_phase_comp(b, p, j):
        raise NotImplementedError(_msg(b, "cp_mol_phase_comp"))

    @staticmethod
    def cv_mol_phase(b, p):
        raise NotImplementedError(_msg(b, "cv_mol_phase"))

    @staticmethod
    def cv_mol_phase_comp(b, p, j):
        raise NotImplementedError(_msg(b, "cv_mol_phase_comp"))

    @staticmethod
    def cv_mol_ig_comp_pure(b, j):
        # Method for calculating pure component ideal gas cv from cp
        # For ideal gases, cv = cp - R
        units = b.params.get_metadata().derived_units
        R = pyunits.convert(const.gas_constant,
                            to_units=units["heat_capacity_mole"])
        return (get_method(b, "cp_mol_ig_comp", j)(
            b, cobj(b, j), b.temperature) - R)

    @staticmethod
    def cv_mol_ls_comp_pure(b, j):
        # Method for calculating pure component liquid and solid cv from cp
        # For ideal (incompressible) liquids and solids, cv = cp
        return get_method(b, "cp_mol_liq_comp", j)(
            b, cobj(b, j), b.temperature)

    @staticmethod
    def dens_mass_phase(b, p):
        raise NotImplementedError(_msg(b, "dens_mass_phase"))

    @staticmethod
    def dens_mol_phase(b, p):
        raise NotImplementedError(_msg(b, "dens_mol_phase"))

    @staticmethod
    def energy_internal_mol_phase(b, p):
        raise NotImplementedError(_msg(b, "energy_internal_mol_phase"))

    @staticmethod
    def energy_internal_mol_phase_comp(b, p, j):
        raise NotImplementedError(_msg(b, "energy_internal_mol_phase_comp"))

    @staticmethod
    def energy_internal_mol_ig_comp_pure(b, j):
        # Method for calculating pure component U from H for ideal gases
        units = b.params.get_metadata().derived_units
        R = pyunits.convert(const.gas_constant,
                            to_units=units["heat_capacity_mole"])

        if cobj(b, j).parent_block().config.include_enthalpy_of_formation:
            # First, need to determine correction between U_form and H_form
            # U_form = H_form - delta_n*R*T
            ele_comp = cobj(b, j).config.elemental_composition
            if ele_comp is None:
                raise ConfigurationError(
                    "{} calculation of internal energy requires elemental "
                    "composition of all species. Please set this using the "
                    "elemental_composition argument in the component "
                    "declaration ({}).".format(b.name, j))

            delta_n = 0
            for e, s in ele_comp.items():
                # Check for any element which is vapor at standard state
                if e in ["He", "Ne", "Ar", "Kr", "Xe", "Ra"]:
                    delta_n += -s
                elif e in ["F", "Cl", "H", "N", "O"]:
                    delta_n += -s/2  # These are diatomic at standard state

            delta_n += 1  # One mole of gaseous compound is formed
            dU_form = delta_n*R*b.params.temperature_ref
        else:
            dU_form = 0  # No heat of formation to correct

        # For ideal gases, U = H - R(T-T_ref) + dU_form
        return (get_method(b, "enth_mol_ig_comp", j)(
            b, cobj(b, j), b.temperature) -
            R*(b.temperature-b.params.temperature_ref) +
            dU_form)

    @staticmethod
    def energy_internal_mol_ls_comp_pure(b, j):
        # Method for calculating pure component U from H for liquids & solids
        units = b.params.get_metadata().derived_units
        R = pyunits.convert(const.gas_constant,
                            to_units=units["heat_capacity_mole"])

        if cobj(b, j).parent_block().config.include_enthalpy_of_formation:
            # First, need to determine correction between U_form and H_form
            # U_form = H_form - delta_n*R*T
            ele_comp = cobj(b, j).config.elemental_composition
            if ele_comp is None:
                raise ConfigurationError(
                    "{} calculation of internal energy requires elemental "
                    "composition of all species. Please set this using the "
                    "elemental_composition argument in the component "
                    "declaration ({}).".format(b.name, j))

            delta_n = 0
            for e, s in ele_comp.items():
                # Check for any element which is vapor at standard state
                if e in ["He", "Ne", "Ar", "Kr", "Xe", "Ra"]:
                    delta_n += -s
                elif e in ["F", "Cl", "H", "N", "O"]:
                    delta_n += -s/2  # These are diatomic at standard state
            dU_form = delta_n*R*b.params.temperature_ref

            # For ideal (incompressible) liquids and solids, U = H + dU_form
            return (get_method(b, "enth_mol_liq_comp", j)(
                b, cobj(b, j), b.temperature) +
                dU_form)
        else:
            # If not including heat of formation, U = H
            return get_method(b, "enth_mol_liq_comp", j)(
                b, cobj(b, j), b.temperature)

    @staticmethod
    def enth_mol_phase(b, p):
        raise NotImplementedError(_msg(b, "enth_mol_phase"))

    @staticmethod
    def enth_mol_phase_comp(b, p, j):
        raise NotImplementedError(_msg(b, "enth_mol_phase_comp"))

    @staticmethod
    def entr_mol_phase(b, p):
        raise NotImplementedError(_msg(b, "entr_mol_phase"))

    @staticmethod
    def entr_mol_phase_comp(b, p, j):
        raise NotImplementedError(_msg(b, "entr_mol_phase_comp"))

    @staticmethod
    def fug_phase_comp(b, p, j):
        raise NotImplementedError(_msg(b, "fug_phase_comp"))

    @staticmethod
    def fug_phase_comp_eq(b, p, j, pp):
        raise NotImplementedError(_msg(b, "fug_phase_comp_eq"))

    @staticmethod
    def log_fug_phase_comp_eq(b, p, j, pp):
        raise NotImplementedError(_msg(b, "log_fug_phase_comp_eq"))

    @staticmethod
    def fug_coeff_phase_comp(b, p, j):
        raise NotImplementedError(_msg(b, "fug_coeff_phase_comp"))

    @staticmethod
    def fug_coeff_phase_comp_eq(b, p, j, pp):
        raise NotImplementedError(_msg(b, "fug_coeff_phase_comp_eq"))

    @staticmethod
    def fug_phase_comp_Tbub(b, p, j, pp):
        raise NotImplementedError(_msg(b, "fug_phase_comp_Tbub"))

    @staticmethod
    def fug_phase_comp_Tdew(b, p, j, pp):
        raise NotImplementedError(_msg(b, "fug_phase_comp_Tdew"))

    @staticmethod
    def fug_phase_comp_Pbub(b, p, j, pp):
        raise NotImplementedError(_msg(b, "fug_phase_comp_Pbub"))

    @staticmethod
    def fug_phase_comp_Pdew(b, p, j, pp):
        raise NotImplementedError(_msg(b, "fug_phase_comp_Pdew"))

    @staticmethod
    def gibbs_mol_phase(b, p):
        raise NotImplementedError(_msg(b, "gibbs_mol_phase"))

    @staticmethod
    def gibbs_mol_phase_comp(b, p, j):
        raise NotImplementedError(_msg(b, "gibbs_mol_phase_comp"))


def _msg(b, attr):
    return ("{} Equation of State module has not implemented a method for {}. "
            "Please contact the EoS developer or use a different module."
            .format(b.name, attr))



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
import numpy as np

from ..atom import Atom


def readPointsFromXYZ(filename):
    Points = []
    with open(filename, 'r') as infile:
        nAtoms = int(infile.readline().split()[0])
        _ = next(infile)  # skip the comment line
        for _ in range(nAtoms):
            line = next(infile).split()
            x = float(line[1])
            y = float(line[2])
            z = float(line[3])
            Points.append(np.array([x, y, z], dtype=float))
    return Points


def readAtomsFromXYZ(filename):
    Atoms = []
    with open(filename, 'r') as infile:
        nAtoms = int(infile.readline().split()[0])
        _ = next(infile)  # skip the comment line
        for _ in range(nAtoms):
            line = next(infile).split()
            Atoms.append(Atom(line[0]))
    return Atoms


def readPointsAndAtomsFromXYZ(filename):
    return readPointsFromXYZ(filename), readAtomsFromXYZ(filename)


def writeDesignToXYZ(D, filename, comment_line=None):
    with open(filename, 'w') as outfile:
        outfile.write('{:d}\n'.format(D.NonVoidCount))  # number of atoms
        outfile.write('{}\n'.format(comment_line if comment_line is not None else ''))
        for i in range(len(D)):
            if not (D.Contents[i] is None or D.Contents[i] == Atom()):
                outfile.write('{} {:.8f} {:.8f} {:.8f}\n'.format(D.Contents[i].Symbol,
                                                                 D.Canvas.Points[i][0],
                                                                 D.Canvas.Points[i][1],
                                                                 D.Canvas.Points[i][2]))



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Flowsheet-related classes and functions used by the UI.
"""
# stdlib
from collections import defaultdict
import copy
import json
import re
from typing import Dict, List, Tuple

# third-party
import pandas as pd
import numpy as np
from pyomo.environ import Block, value
from pyomo.network import Arc
from pyomo.network.port import Port

# package
from idaes import logger
from idaes.ui.icons import UnitModelIcon

_log = logger.getLogger(__name__)


class FileBaseNameExistsError(Exception):
    pass


def validate_flowsheet(fs: Dict) -> Tuple[bool, str]:
    """Validate a flowsheet.

    Expected format is below.

    .. code-block:: json

        {
            "model": {
                "id": "<model name>",
                "unit_models": {
                    "<component name>": {
                        "image": "<image name>",
                        "type": "<component type name>",
                        "...": "more values..."
                    },
                    "...": "etc."
                },
                "arcs": {
                    "<arc name>": {
                        "source": "<component name>",
                        "dest": "<component name>",
                        "label": "<label text>"
                    },
                    "...": "etc."
                }
            },
            "cells": [
                {"id": "<component_name>", "...": "other values used by JointJS.."},
                "..."
            ]
        }

    Args:
        fs: Flowsheet to validate

    Return:
        Tuple of (True, "") for OK, and (False, "<message>") for failure
    """
    # very quick and dirty validation, but it does make for nice clean error messages
    for key in "model", "cells":
        if key not in fs:
            return False, f"Missing top-level key '{key}'"
    model, component_ids = fs["model"], set()
    for key2 in "id", "unit_models", "arcs":
        if key2 not in model:
            return False, f"The flowsheet model is missing key '{key2}'"
        if key2 == "unit_models":
            for ckey, cval in model[key2].items():
                for key3 in "image", "type":
                    if key3 not in cval:
                        return False, f"Unit model '{ckey}' is missing key '{key3}'"
                component_ids.add(ckey)
        elif key2 == "arcs":
            for akey, aval in model[key2].items():
                for key3 in "source", "dest", "label":
                    if key3 not in aval:
                        return False, f"Arc '{akey}' is missing key '{key3}'"
                component_ids.add(akey)
    cells = fs["cells"]
    # Check if all cell id's are in the model
    cell_ids = set()
    for i, cell in enumerate(cells):
        if "id" not in cell:
            return False, f"Cell #{i + 1} is missing key 'id'"
        cell_id = cell["id"]
        if cell_id not in component_ids:
            return False, f"Cell id '{cell_id}' not found in unit models or arcs"
        cell_ids.add(cell_id)
    # Check if all model id's are in the cells
    if cell_ids != component_ids:
        missing = component_ids - cell_ids
        sfx = "s" if len(missing) > 1 else ""
        return False, f"Component id{sfx} {missing} {'are' if sfx else 'is'} not in the layout cells"
    return True, ""


class FlowsheetSerializer:
    """Serializes the flowsheet into one dict with two sections.

    The "model" section contains the id of the flowsheet and the
    unit models and arcs. This will be used to compare the model and convert
    to jointjs. The "cells" section is the jointjs readable code.
    See :py:func:`validate_flowsheet` for details on the format.
    """
    #: Regular expression identifying inlets by last component of ports' name
    INLET_REGEX = re.compile(
        r"^in_|^feed_|^inlet_|^in$|^feed$|^inlet$|_in$|_feed$|_inlet$", re.IGNORECASE
    )
    #: Regular expression identifying outlets by last component of ports' name
    OUTLET_REGEX = re.compile(
        r"^out_|^prod_|^outlet_|^out$|^prod$|^outlet$|_out$|_prod$|_outlet$",
        re.IGNORECASE,
    )

    def __init__(self, flowsheet, name: str, validate: bool = True):
        """Serialize input flowsheet with given name

        Args:
            flowsheet: The flowsheet to serialize
            name: The name of the flowsheet (also called its 'id' in some contexts)
            validate: If True, validate that the flowsheet is a reaonsable IDAES model, first
        Raises:
            ValueError if validation is on and flowsheet is found to be invalid
        """
        # validate, if so directed
        if validate:
            if not hasattr(flowsheet, "component_objects") or not callable(flowsheet.component_objects):
                raise ValueError("Flowsheet missing function 'component_objects'")
            try:
                n_obj = 0
                for objtype in Arc, Block:
                    for obj in flowsheet.component_objects(objtype, descend_into=False):
                        n_obj += 1
                        if not hasattr(obj, "getname") or not callable(obj.getname):
                            raise ValueError("Flowsheet component missing function 'getname'")
            except ValueError:
                raise
            except Exception as err:
                raise ValueError(f"Flowsheet 'component_objects' cannot be navigated: {err}")
            if n_obj == 0:
                raise ValueError("Flowsheet has no Arcs or unit Blocks")
        # setup
        self.unit_models = {}  # {unit: {"name": unit.getname(), "type": str?}}
        self.arcs = {}  # {Arc.getname(): Arc}
        self.ports = {}  # {Port: parent_unit}
        self.edges = defaultdict(list)  # {name: {"source": unit, "dest": unit}}
        self.orphaned_ports = {}
        self.labels = {}
        self._stream_table_df = None
        self._out_json = {"model": {}}
        self._serialized_contents = defaultdict(dict)
        self._used_ports = set()
        self._known_endpoints = set()
        self._unit_name_used_count = defaultdict(lambda: 0)
        self._sig_figs = 5  # Defines the number of significant figures after the decimal place
        self._logger = logger.getLogger(__name__)
        self.name = name
        self.flowsheet = flowsheet
        # serialize
        self._ingest_flowsheet()
        self._construct_output_json()

    def as_dict(self):
        return self._out_json

    def _ingest_flowsheet(self):
        # Stores information on the connectivity and components of the input flowsheet
        self._identify_arcs()
        self._identify_unit_models()
        self._construct_stream_labels()
        untouched_ports = self._map_edges()
        self._identify_implicit_feeds_and_products(untouched_ports)

    def _identify_arcs(self):
        # Identify the arcs and known endpoints and store them
        for component in self.flowsheet.component_objects(Arc, descend_into=False):
            self.arcs[component.getname()] = component
            self._known_endpoints.add(component.source.parent_block())
            self._known_endpoints.add(component.dest.parent_block())

    def _identify_unit_models(self):
        from idaes.core import UnitModelBlockData  # avoid circular import

        # Identify the unit models and ports and store them
        for component in self.flowsheet.component_objects(Block, descend_into=True):
            if isinstance(component, UnitModelBlockData):
                self._add_unit_model_with_ports(component)
            else:
                # some unit models are nested within Indexed blocks
                component_object_op = getattr(component, "component_object", None)
                if not callable(component_object_op):
                    for item in component.parent_component().values():
                        if isinstance(item, UnitModelBlockData):
                            if any(
                                (
                                    item == arc.source.parent_block()
                                    or item == arc.dest.parent_block()
                                )
                                for arc in self.arcs.values()
                            ):
                                self._add_unit_model_with_ports(item)

    def _construct_stream_labels(self):
        # Construct the stream labels
        from idaes.core.util.tables import (
            stream_states_dict,
        )  # deferred to avoid circ. import

        # We might have this information from generating self.serialized_components but I (Makayla) don't
        # know how that connects to the stream names so this will be left alone for now
        for stream_name, stream_value in stream_states_dict(self.arcs).items():
            label = ""
            for var, var_value in stream_value.define_display_vars().items():
                var = var.capitalize()

                for k, v in var_value.items():
                    if k is None:
                        label += f"{var} {round(value(v), self._sig_figs)}\n"
                    else:
                        label += f"{var} {k} {round(value(v), self._sig_figs)}\n"
            self.labels[stream_name] = label[:-2]

    def _map_edges(self):
        # Map the arcs to the ports to construct the edges
        used_ports = set()
        for name, arc in self.arcs.items():
            try:  # This is necessary because for internally-nested arcs we may not record ports
                self.edges[name] = {
                    "source": self.ports[arc.source],
                    "dest": self.ports[arc.dest],
                }
                used_ports.add(arc.source)
                used_ports.add(arc.dest)
            except KeyError as error:
                self._logger.error(
                    f"Unable to find port. {name}, {arc.source}, {arc.dest}"
                )

        # Note we're only using the keys from self.ports, {port: parentcomponent}
        return set(self.ports) - used_ports

    def _make_valid_json(self, df):
        """
        Replacing NaN, Infinity, and Negative Infinity with strings to make valid JSON for front end consumption
        Args:
            df: The Pandas dataframe to replace NaN, Infinity, and Negative Infinity with strings
        Return:
            The dataframe that now has valid JSON
        """
        df = df.replace(np.nan, "NaN").replace(np.inf, "Inf").replace(np.NINF, "-Inf")
        return df

    def _add_unit_model_with_ports(self, unit):
        unit_name = unit.getname()
        if unit.parent_block() == self.flowsheet:
            # The unit is top-level and therefore should be displayed.
            self.unit_models[unit] = {
                "name": unit_name,
                "type": self._get_unit_model_type(unit),
            }

            self._unit_name_used_count[unit_name] += 1
            for port in unit.component_objects(Port, descend_into=False):
                self.ports[port] = unit
            

            performance_contents, stream_df = unit.serialize_contents()
            if stream_df is not None and not stream_df.empty:
                # If there is a stream dataframe then we need to reset the index so we can get the variable names
                # and then rename the "index"
                stream_df = stream_df.reset_index().rename(
                    columns={"index": "Variable"}
                )
                stream_df = self._make_valid_json(stream_df)
            self._serialized_contents[unit_name]["stream_contents"] = stream_df

            performance_df = pd.DataFrame()
            if performance_contents:
                # If performance contents is not empty or None then stick it into a dataframe and convert the
                # GeneralVars to actual values
                performance_df = pd.DataFrame(
                    performance_contents["vars"].items(), columns=["Variable", "Value"]
                )
                performance_df["Value"] = performance_df["Value"].map(value)
                performance_df = self._make_valid_json(performance_df)
            self._serialized_contents[unit_name]["performance_contents"] = performance_df

        elif unit in self._known_endpoints:
            # Unit is a subcomponent AND it is connected to an Arc. Or maybe it's in an indexed block TODO CHECK
            # Find the top-level parent unit and assign the serialized link to the parent.
            parent_unit = unit.parent_block()
            while not parent_unit == self.flowsheet:
                parent_unit = parent_unit.parent_block()

            self.unit_models[parent_unit] = {
                "name": parent_unit.getname(),
                "type": self._get_unit_model_type(parent_unit),
            }
            for port in unit.component_objects(Port, descend_into=False):
                self.ports[port] = parent_unit
                # Add all of this subcomponent's ports to used_ports; realistically, this is only relevant
                # in a situation where the subcomponent has ports that are not connected to an Arc that we
                # intend to display. Otherwise, the port is marked as "used" when we traverse the Arcs.
                self._used_ports.add(port)
        else:
            # The unit is neither top-level nor connected; do not display this unit, since it is a subcomponent.
            pass

    def _get_unit_model_type(self, unit):
        # Get the unit models type
        return unit.base_class_module().split(".")[
            -1
        ]  # TODO look for/create equivalent getter, as getname() above

    def _identify_implicit_feeds_and_products(self, untouched_ports):
        # Identify feeds and products not explicitly defined by the user by examining the names of ports
        # not connected to arcs
        # This is intended for use on ports of top level unit models. It is unclear if it works with nested
        # unit models (probably not)
        for port in sorted(
            untouched_ports, key=lambda port: str(port)
        ):  # Sorting for determinism
            port_name = str(port).split(".")[-1]
            unit_name = self._unique_unit_name(
                port_name
            )  # this becomes the I/O's ID and label
            edge_name = f"s_{unit_name}"
            # identify ports per INLET_REGEX/OUTLET_REGEX
            # then pretend they're unit models
            # then add their edges
            feed_match = self.INLET_REGEX.search(port_name)
            if feed_match:
                # name the feed "unit model" and its connecting edge with the name of the port itself
                feed_port = self._PseudoUnit("Feed", unit_name)
                self.unit_models[feed_port] = {
                    "name": feed_port.getname(),
                    "type": "feed",
                }
                self.edges[edge_name] = {"source": feed_port, "dest": self.ports[port]}
                arc_label = "feed info"
                self.labels[
                    edge_name
                ] = arc_label  # We want the arc label to look identical to that of a user defined feed
                continue

            product_match = self.OUTLET_REGEX.search(port_name)
            if product_match:
                prod_port = self._PseudoUnit("Product", unit_name)
                self.unit_models[prod_port] = {
                    "name": prod_port.getname(),
                    "type": "product",
                }
                self.edges[edge_name] = {"source": self.ports[port], "dest": prod_port}
                arc_label = "product info"
                self.labels[
                    edge_name
                ] = arc_label  # We want the arc label to look identical to that of a user defined product
                continue

            if not feed_match and not product_match:
                # Just in case there is a disconnected port that doesn't match the regexes display a warning
                self._logger.warning(
                    f"Disconnected port found: {port_name} parent unit model: {self.ports[port].getname()}"
                )

    def _unique_unit_name(self, base_name):
        # Prevent name collisions by simply appending a suffix based on how many times the name's been used
        self._unit_name_used_count[base_name] += 1
        return f"{base_name}_{self._unit_name_used_count[base_name]}"

    def _construct_output_json(self):
        self._construct_model_json()
        self._construct_jointjs_json()

    def _construct_model_json(self):
        from idaes.core.util.tables import create_stream_table_dataframe # deferred to avoid circular import

        # Get the stream table and add it to the model json
        # Change the index of the pandas dataframe to not be the variables
        self._stream_table_df = (
            create_stream_table_dataframe(self.arcs)
                # Change the index of the pandas dataframe to not be the variables
            .reset_index().rename(columns={"index": "Variable"})
            .reset_index().rename(columns={"index": ""})
            .applymap(lambda x: round(x, self._sig_figs) if isinstance(x, (int, float)) else x)
        )
        
        # Change NaNs to None for JSON
        self._stream_table_df = self._stream_table_df.where((pd.notnull(self._stream_table_df)), None)

        # Puts df in this format for easier parsing in the javascript table:
        # {'index': ["('Liq', 'benzene')", "('Liq', 'toluene')", "('Liq', 'hydrogen')", "('Liq', 'methane')", "('Vap', 'benzene')", "('Vap', 'toluene')", "('Vap', 'hydrogen')", "('Vap', 'methane')", 'temperature', 'pressure'], 
        # 'columns': ['s03', 's04', 's05', 's06', 's08', 's09', 's10'], 
        # 'data': [[0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5], [0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5], [0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5], [0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5], [0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5], [0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5], [0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5], [0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5], [298.15, 298.15, 298.15, 298.15, 298.15, 298.15, 298.15], [101325.0, 101325.0, 101325.0, 101325.0, 101325.0, 101325.0, 101325.0]]}
        self._out_json["model"]["stream_table"] = self._stream_table_df.to_dict("split")

        self._out_json["model"]["id"] = self.name
        self._out_json["model"]["unit_models"] = {}
        self._out_json["model"]["arcs"] = {}

        for unit_model in self.unit_models.values():
            unit_name = unit_model["name"]
            unit_type = unit_model["type"]
            unit_icon = UnitModelIcon(unit_type)

            unit_contents = {
                "type": unit_type,
                "image": "/images/icons/" + unit_icon.icon,
            }
            if unit_name in self._serialized_contents:
                for pfx in "performance", "stream":
                    content_type = pfx + "_contents"
                    c = self._serialized_contents[unit_name][content_type].applymap(lambda x: round(x, self._sig_figs) if isinstance(x, (int, float)) else x).to_dict("index")
                    # ensure that keys are strings (so it's valid JSON)
                    unit_contents[content_type] = {str(k): v for k, v in c.items()}

            self._out_json["model"]["unit_models"][unit_name] = unit_contents

        for edge, edge_info in self.edges.items():
            self._out_json["model"]["arcs"][edge] = {
                "source": edge_info["source"].getname(),
                "dest": edge_info["dest"].getname(),
                "label": self.labels[edge],
            }

    def _construct_jointjs_json(self):
        self._out_json["cells"] = []

        # Start out in the top left corner until we get a better inital layout
        x_pos = 10
        y_pos = 10
        y_starting_pos = 10

        default_icon = UnitModelIcon()
        for component, unit_attrs in self.unit_models.items():
            unit_icon = UnitModelIcon(unit_attrs["type"])
            try:
                self._create_image_jointjs_json(
                    x_pos,
                    y_pos,
                    unit_attrs["name"],
                    unit_icon.icon,
                    unit_attrs["type"],
                    unit_icon.link_positions,
                )
            except KeyError as e:
                self._logger.info(
                    f'Unable to find icon for {unit_attrs["type"]}. Using default icon'
                )
                self._create_image_jointjs_json(
                    x_pos,
                    y_pos,
                    unit_attrs["name"],
                    default_icon.icon,
                    unit_attrs["type"],
                    default_icon.link_positions,
                )

            # If x_pos it greater than 700 then start another diagonal line
            if x_pos >= 700:
                x_pos = 100
                y_pos = y_starting_pos
                y_starting_pos += 100
            else:
                x_pos += 100
                y_pos += 100

        for name, ports_dict in self.edges.items():
            umst = self.unit_models[ports_dict["source"]]["type"]  # alias
            dest = ports_dict["dest"]
            if hasattr(ports_dict["source"], "vap_outlet"):
                # TODO Figure out how to denote different outlet types. Need to
                # deal with multiple input/output offsets
                for arc in list(self.arcs.values()):
                    if (
                        self.ports[arc.dest] == dest
                        and arc.source == ports_dict["source"].vap_outlet
                    ):
                        source_anchor = "top"
                    else:
                        source_anchor = "bottom"
            else:
                source_anchor = "out"

            # The source_port and dest_port should be replaced by actual names in case there are multiple
            # inlets and outlets between the same two unit models
            source_port = "out"
            dest_port = "in"
            self._create_link_jointjs_json(
                source_port,
                dest_port,
                ports_dict["source"].getname(),
                dest.getname(),
                name,
                self.labels[name],
            )

    def _create_image_jointjs_json(
        self, x_pos, y_pos, name, image, title, port_groups
    ):
        # Create the jointjs for a given image
        entry = {}
        entry["type"] = "standard.Image"
        entry["position"] = {"x": x_pos, "y": y_pos}
        # The icon width and height default to 50x50 making all icons a square. This will need to be changed
        # when we have more unit models that should not be square. Probaly add it to the icon mapping
        icon_width = 50
        icon_height = 50
        entry["size"] = {"width": icon_width, "height": icon_height}
        # We want the icons to not be at an angle initially
        angle = 0
        entry["angle"] = angle
        entry["id"] = name
        # This defines what layer the icon is on
        z = 1
        entry["z"] = z
        entry["ports"] = port_groups
        entry["attrs"] = {
            "image": {"xlinkHref": "/images/icons/" + image},
            "label": {"text": name},
            "root": {"title": title},
        }
        self._out_json["cells"].append(entry)

    def _create_link_jointjs_json(
        self, source_port, dest_port, source_id, dest_id, name, label
    ):
        # Create the joint js for a link
        # Set the padding to 10. Makayla saw it in a jointjs example
        padding = 10
        # Set the initial offset position for the link labels. Makayla saw these numbers in a jointjs example
        position_distance = 0.66
        position_offset = -40
        z = 2
        entry = {
            "type": "standard.Link",
            "source": {"id": source_id, "port": source_port},
            "target": {"id": dest_id, "port": dest_port},
            "router": {"name": "orthogonal", "padding": padding},
            "connector": {"name": "jumpover", "attrs": {"line": {"stroke": "#5c9adb"}}},
            "id": name,
            "labels": [
                # This label MUST be first or the show/hide will fail
                {
                    "attrs": {
                        # Start with the labels off
                        "rect": {
                            "fill": "#d7dce0",
                            "stroke": "white",
                            "stroke-width": 0,
                            "fill-opacity": 0,
                        },
                        "text": {
                            "text": label,
                            "fill": "black",
                            "text-anchor": "left",
                            "display": "none",
                        },
                    },
                    "position": {
                        "distance": position_distance,
                        "offset": position_offset,
                    },
                },
                {"attrs": {"text": {"text": name}}},
            ],
            "z": z,
        }
        self._out_json["cells"].append(entry)

    class _PseudoUnit:
        """
        Unit-like object in order to emulate missing unit models for implicit feed/product
        eventually may need to actually implement/subclass
        """

        def __init__(self, typename, id):
            self.name = typename
            self.id = id

        def getname(self):
            return self.id


class FlowsheetDiff:
    """Compute a flowsheet model 'diff' and use that to compute an updated layout.

    Flowsheets are serialized by the core library. See :func:`validate_flowsheet` for the required format.

    Example usage::

        diff =  FlowsheetDiff(saved_flowsheet, new_flowsheet)
        merged_flowsheet = {"model": new_flowsheet["model"], "cells": diff.layout}
    """
    def __init__(self, old_flowsheet: Dict, new_flowsheet: Dict, validate=True):
        """Construct with old and new flowsheet, and compute the diff and new layout

        Args:
            old_flowsheet: Old flowsheet value
            new_flowsheet: New flowsheet value
            validate: If True, validate each flowsheet before doing the diff
        Raises:
            ValueError if either flowsheet is invalid
        """
        if validate:
            ok, why = validate_flowsheet(old_flowsheet)
            if not ok:
                raise ValueError(f"invalid old_flowsheet value: {why}")
            ok, why = validate_flowsheet(new_flowsheet)
            if not ok:
                raise ValueError(f"invalid new_flowsheet value: {why}")
        self._old, self._new = old_flowsheet, new_flowsheet
        self._len = None
        self._diff = self._compute_diff()
        self._layout = self._compute_layout() if self._diff else None

    def merged(self, do_copy: bool = False) -> Dict:
        """Return merged flowsheet.

        If the diff is empty, this will be the 'old_flowsheet' (or a copy) passed to the constructor.

        Args:
            do_copy: If True, return a copy so modifying the returned dict won't affect input arguments.

        Returns:
            Merged dict with structure as described in :func:`validate_flowsheet()`.
        """
        if bool(self):
            result = {
                "model": copy.deepcopy(self._new["model"]) if do_copy else self._new["model"],
                "cells": self._layout
            }
        else:
            # diff is empty, return 'old' object
            result = copy.deepcopy(self._old) if do_copy else self._old
        return result

    def __len__(self):
        if self._len is None:
            return 0
        return self._len

    def __bool__(self):
        return len(self) > 0

    def __str__(self):
        return json.dumps(self._diff, indent=2)

    def _compute_diff(self) -> Dict:
        diff = {"add": {}, "remove": {}, "change": {}}
        old_model, new_model = self._old["model"], self._new["model"]
        n = 0
        for cls in "unit_models", "arcs":
            for k in diff.keys():
                diff[k][cls] = {}
            old_data, new_data = old_model[cls], new_model[cls]
            # Add/change
            for key in new_data:
                if key not in old_data:
                    diff["add"][cls][key] = copy.deepcopy(new_data[key])
                    n += 1
                elif old_data[key] != new_data[key]:
                    diff["change"][cls][key] = copy.deepcopy(new_data[key])
                    n += 1
            # Remove
            for key in old_data:
                if key not in new_data:
                    diff["remove"][cls][key] = True
                    n += 1
        self._len = n
        return diff

    def _compute_layout(self) -> List:
        """Based on the old and new flowsheets provided in the constructor, compute the layout
        that should be placed in the "cells" of the new flowsheet.

        Returns:
            New layout to put into the "cells" of the new flowsheet dict
        """
        layout, x, y = [], 50, 50
        # Add
        for cls in self._diff["add"]:
            for id_ in self._diff["add"][cls]:
                values = self._diff["add"][cls][id_]
                if cls == "arcs":
                    new_item = self._new_arc(id_, values)
                else:
                    new_item = self._new_unit_model(id_, values, x, y)
                    x, y = x + 100, y + 100
                layout.append(new_item)
        # Change, remove, and simply copy
        for item in self._old["cells"]:
            id_ = item["id"]
            cls = "arcs" if "source" in item else "unit_models"
            if id_ in self._diff["remove"][cls]:
                continue
            elif id_ in self._diff["change"][cls]:
                values = self._diff["change"][cls][id_]
                if cls == "arcs":
                    new_item = self._update_arc(item, values)
                else:
                    new_item = self._update_unit_model(item, values)
                layout.append(new_item)
            else:
                layout.append(copy.deepcopy(item))
        return layout

    @staticmethod
    def _new_arc(name, values):
        return {
            "type": "standard.Link",
            "source": {"id": values["source"]},
            "target": {"id": values["dest"]},
            "router": {"name": "orthogonal", "padding": 10},
            "connector": {"name": "normal", "attrs": {"line": {"stroke": "#5c9adb"}}},
            "id": name,
            "labels": [
                {
                    "attrs": {
                        "rect": {"fill": "#d7dce0", "stroke": "#FFFFFF", "stroke-width": 1},
                        "text": {
                            "text": values["label"],
                            "fill": "black",
                            "text-anchor": "left",
                        },
                    },
                    "position": {"distance": 0.66, "offset": -40},
                }
            ],
            "z": 2,
        }

    @staticmethod
    def _update_arc(item, values):
        new_item = copy.deepcopy(item)
        new_item["source"]["id"] = values["source"]
        new_item["target"]["id"] = values["dest"]
        new_item["labels"][0]["attrs"]["text"]["text"] = values["label"]
        return new_item

    @staticmethod
    def _new_unit_model(name, values, x, y):
        return {
            "type": "standard.Image",
            "id": name,
            "position": {"x": x, "y": y},
            "size": {"width": 50, "height": 50},
            "angle": 0,
            "z": 1,
            "attrs": {
                "image": {"xlinkHref": values["image"]},
                "label": {"text": name},
                "root": {"title": values["type"]},
            },
        }

    @staticmethod
    def _update_unit_model(item, values):
        new_item = copy.deepcopy(item)
        new_item["attrs"]["image"]["xlinkHref"] = values["image"]
        new_item["attrs"]["root"]["title"] = values["type"]
        return new_item





##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
import pyomo.environ as pyo
import pytest
import idaes
import idaes.core.solvers as isolve

@pytest.mark.unit
def test_ipopt_idaes_available():
    """
    Tries to set-up the IPOPT with the IDAES SolverFactory wrapper
    """
    if not pyo.SolverFactory('ipopt').available():
        raise Exception(
            "Could not find IPOPT. Users are strongly encouraged to have a "
            "version of IPOPT available, as it is the default solver assumed "
            "by many IDAES examples and tests. See the IDAES install "
            "documentation for instructions on how to get IPOPT.")

@pytest.mark.skipif(not pyo.SolverFactory('ipopt').available(False), reason="no Ipopt")
@pytest.mark.unit
def test_ipopt_idaes_config():
    """
    Test that the default solver options are set
    """
    with idaes.temporary_config_ctx():
        # in this context use idaes default solver options
        isolve.use_idaes_solver_configuration_defaults()
        idaes.cfg.ipopt.options.nlp_scaling_method = "toast-based"
        solver = pyo.SolverFactory('ipopt')
        assert solver.options["nlp_scaling_method"] == "toast-based"
        solver = pyo.SolverFactory('ipopt', options={"tol":1})
        assert solver.options["tol"] == 1
        idaes.cfg.ipopt.options.tol = 1
        solver = pyo.SolverFactory('ipopt')
        assert solver.options["tol"] == 1
    #
    # TODO(JCE): Bring back the rest of this test after the tests
    # untangled in the GT PR
    #
    # back to the original config, don't use idaes default option settings
    #solver = pyo.SolverFactory('ipopt')
    # should not be using idaes defaults here so options should be empty
    # if this fails there is a very good chance another test was set to use
    # idaes defaults.
    #assert "nlp_scaling_method" not in solver.options

@pytest.mark.skipif(not pyo.SolverFactory('ipopt').available(False), reason="no Ipopt")
@pytest.mark.unit
def test_ipopt_idaes_solve():
    """
    Make sure there is no issue with the solver class or default settings that
    break the solver object.  Passing a bad solver option will result in failure
    """
    solver = pyo.SolverFactory('ipopt')
    m = pyo.ConcreteModel()
    m.x = pyo.Var(initialize=-0.1)
    m.y = pyo.Var(initialize=1)
    m.c1 = pyo.Constraint(expr=1==m.x**3)
    m.c2 = pyo.Constraint(expr=1==m.y)
    solver.solve(m)
    assert pytest.approx(1) == pyo.value(m.x)

@pytest.mark.skipif(not pyo.SolverFactory('ipopt').available(False), reason="no Ipopt")
@pytest.mark.unit
def test_default_solver():
    """Test that default solver returns the correct solver type
    """
    assert type(pyo.SolverFactory("default")) \
        == type(pyo.SolverFactory(idaes.cfg.default_solver))



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for Pressure Changer unit model.

Author: Andrew Lee, Emmanuel Ogbe
"""
import pytest

from pyomo.environ import (ConcreteModel,
                           Constraint,
                           TerminationCondition,
                           SolverStatus,
                           units,
                           value,
                           Var)
from pyomo.util.check_units import (assert_units_consistent,
                                    assert_units_equivalent)
from pyomo.util.calc_var_value import calculate_variable_from_constraint

from idaes.core import (FlowsheetBlock,
                        MaterialBalanceType,
                        EnergyBalanceType,
                        MomentumBalanceType)

from idaes.generic_models.unit_models.pressure_changer import (
    PressureChanger,
    PressureChangerData,
    Turbine,
    Compressor,
    Pump,
    ThermodynamicAssumption,
)

from idaes.generic_models.properties.activity_coeff_models.BTX_activity_coeff_VLE \
    import BTXParameterBlock
from idaes.generic_models.properties import iapws95
from idaes.generic_models.properties.examples.saponification_thermo import \
    SaponificationParameterBlock

from idaes.core.util.model_statistics import (degrees_of_freedom,
                                              number_variables,
                                              number_total_constraints,
                                              number_unused_variables)
from idaes.core.util.testing import (PhysicalParameterTestBlock,
                                     initialization_tester)
from idaes.core.util.exceptions import BalanceTypeNotSupportedError
from idaes.core.util import get_solver, scaling as iscale


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


# -----------------------------------------------------------------------------
@pytest.mark.unit
def test_ThermodynamicAssumption():
    assert len(ThermodynamicAssumption) == 4


class TestPressureChanger(object):
    @pytest.mark.unit
    def test_config(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = PhysicalParameterTestBlock()

        m.fs.unit = PressureChanger(default={
                "property_package": m.fs.properties})

        # Check unit config arguments
        assert len(m.fs.unit.config) == 12

        assert m.fs.unit.config.material_balance_type == \
            MaterialBalanceType.useDefault
        assert m.fs.unit.config.energy_balance_type == \
            EnergyBalanceType.useDefault
        assert m.fs.unit.config.momentum_balance_type == \
            MomentumBalanceType.pressureTotal
        assert not m.fs.unit.config.has_phase_equilibrium
        assert m.fs.unit.config.compressor
        assert m.fs.unit.config.thermodynamic_assumption == \
            ThermodynamicAssumption.isothermal
        assert m.fs.unit.config.property_package is m.fs.properties

    @pytest.mark.unit
    def test_dynamic_build(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": True, "time_units": units.s})

        m.fs.properties = PhysicalParameterTestBlock()

        m.fs.unit = PressureChanger(default={
                "property_package": m.fs.properties})
        iscale.calculate_scaling_factors(m)

        assert hasattr(m.fs.unit, "volume")

    @pytest.mark.unit
    def test_pump(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = PhysicalParameterTestBlock()

        m.fs.unit = PressureChanger(default={
                "property_package": m.fs.properties,
                "thermodynamic_assumption": ThermodynamicAssumption.pump})
        iscale.calculate_scaling_factors(m)

        assert isinstance(m.fs.unit.fluid_work_calculation, Constraint)

    @pytest.mark.unit
    def test_adiabatic(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = PhysicalParameterTestBlock()

        m.fs.unit = PressureChanger(default={
                "property_package": m.fs.properties,
                "thermodynamic_assumption": ThermodynamicAssumption.adiabatic})
        iscale.calculate_scaling_factors(m)


    @pytest.mark.unit
    def test_isentropic_comp_phase_balances(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = PhysicalParameterTestBlock()

        m.fs.unit = PressureChanger(default={
                "property_package": m.fs.properties,
                "thermodynamic_assumption": ThermodynamicAssumption.isentropic,
                "material_balance_type": MaterialBalanceType.componentPhase})
        iscale.calculate_scaling_factors(m)

        assert isinstance(m.fs.unit.state_material_balances, Constraint)
        assert len(m.fs.unit.state_material_balances) == 4

    @pytest.mark.unit
    def test_isentropic_comp_total_balances(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = PhysicalParameterTestBlock()

        m.fs.unit = PressureChanger(default={
                "property_package": m.fs.properties,
                "thermodynamic_assumption": ThermodynamicAssumption.isentropic,
                "material_balance_type": MaterialBalanceType.componentTotal})
        iscale.calculate_scaling_factors(m)

        assert isinstance(m.fs.unit.state_material_balances, Constraint)
        assert len(m.fs.unit.state_material_balances) == 2

    @pytest.mark.unit
    def test_isentropic_total_balances(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = PhysicalParameterTestBlock()

        with pytest.raises(BalanceTypeNotSupportedError):
            m.fs.unit = PressureChanger(default={
                "property_package": m.fs.properties,
                "thermodynamic_assumption": ThermodynamicAssumption.isentropic,
                "material_balance_type": MaterialBalanceType.total})

    @pytest.mark.unit
    def test_isentropic_total_element_balances(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = PhysicalParameterTestBlock()

        with pytest.raises(BalanceTypeNotSupportedError):
            m.fs.unit = PressureChanger(default={
                "property_package": m.fs.properties,
                "thermodynamic_assumption": ThermodynamicAssumption.isentropic,
                "material_balance_type": MaterialBalanceType.elementTotal})

    @pytest.mark.unit
    def test_isentropic_material_balances_none(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = PhysicalParameterTestBlock()

        with pytest.raises(BalanceTypeNotSupportedError):
            m.fs.unit = PressureChanger(default={
                "property_package": m.fs.properties,
                "thermodynamic_assumption": ThermodynamicAssumption.isentropic,
                "material_balance_type": MaterialBalanceType.none})


# -----------------------------------------------------------------------------
class TestBTX_isothermal(object):
    @pytest.fixture(scope="class")
    def btx(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = BTXParameterBlock(default={"valid_phase": 'Liq'})

        m.fs.unit = PressureChanger(default={
            "property_package": m.fs.properties,
            "thermodynamic_assumption": ThermodynamicAssumption.isothermal})

        m.fs.unit.inlet.flow_mol[0].fix(5)  # mol/s
        m.fs.unit.inlet.temperature[0].fix(365)  # K
        m.fs.unit.inlet.pressure[0].fix(101325)  # Pa
        m.fs.unit.inlet.mole_frac_comp[0, "benzene"].fix(0.5)
        m.fs.unit.inlet.mole_frac_comp[0, "toluene"].fix(0.5)

        m.fs.unit.deltaP.fix(50000)
        iscale.calculate_scaling_factors(m)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, btx):
        assert hasattr(btx.fs.unit, "inlet")
        assert len(btx.fs.unit.inlet.vars) == 4
        assert hasattr(btx.fs.unit.inlet, "flow_mol")
        assert hasattr(btx.fs.unit.inlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.inlet, "temperature")
        assert hasattr(btx.fs.unit.inlet, "pressure")

        assert hasattr(btx.fs.unit, "outlet")
        assert len(btx.fs.unit.outlet.vars) == 4
        assert hasattr(btx.fs.unit.outlet, "flow_mol")
        assert hasattr(btx.fs.unit.outlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.outlet, "temperature")
        assert hasattr(btx.fs.unit.outlet, "pressure")

        assert hasattr(btx.fs.unit, "work_mechanical")
        assert hasattr(btx.fs.unit, "deltaP")
        assert isinstance(btx.fs.unit.ratioP, Var)
        assert isinstance(btx.fs.unit.ratioP_calculation, Constraint)

        assert number_variables(btx) == 25
        assert number_total_constraints(btx) == 19
        assert number_unused_variables(btx) == 0

    @pytest.mark.component
    def test_units(self, btx):
        assert_units_consistent(btx)
        assert_units_equivalent(btx.fs.unit.work_mechanical[0], units.W)
        assert_units_equivalent(btx.fs.unit.deltaP[0], units.Pa)

    @pytest.mark.unit
    def test_dof(self, btx):
        assert degrees_of_freedom(btx) == 0

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, btx):
        initialization_tester(btx)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, btx):
        results = solver.solve(btx)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, btx):
        assert (pytest.approx(5, abs=1e-3) ==
                value(btx.fs.unit.outlet.flow_mol[0]))
        assert (pytest.approx(365, abs=1e-2) ==
                value(btx.fs.unit.outlet.temperature[0]))
        assert (pytest.approx(151325, abs=1e2) ==
                value(btx.fs.unit.outlet.pressure[0]))
        assert (pytest.approx(0, abs=1e-6) ==
                value(btx.fs.unit.work_mechanical[0]))

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, btx):
        assert abs(value(btx.fs.unit.inlet.flow_mol[0] -
                         btx.fs.unit.outlet.flow_mol[0])) <= 1e-6

        assert abs(value(btx.fs.unit.outlet.flow_mol[0] *
                   (btx.fs.unit.control_volume.properties_in[0]
                    .enth_mol_phase['Liq'] -
                    btx.fs.unit.control_volume.properties_out[0]
                    .enth_mol_phase['Liq']))) <= 1e-6

    @pytest.mark.ui
    @pytest.mark.unit
    def test_report(self, btx):
        btx.fs.unit.report()


# -----------------------------------------------------------------------------
@pytest.mark.iapws
@pytest.mark.skipif(not iapws95.iapws95_available(),
                    reason="IAPWS not available")
class TestIAPWS(object):
    @pytest.fixture(scope="class")
    def iapws(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = iapws95.Iapws95ParameterBlock()

        m.fs.unit = PressureChanger(default={
                "property_package": m.fs.properties,
                "thermodynamic_assumption": ThermodynamicAssumption.isentropic,
                "compressor": True})

        m.fs.unit.inlet.flow_mol[0].fix(100)
        m.fs.unit.inlet.enth_mol[0].fix(4000)
        m.fs.unit.inlet.pressure[0].fix(101325)

        m.fs.unit.deltaP.fix(50000)
        m.fs.unit.efficiency_isentropic.fix(0.9)
        iscale.calculate_scaling_factors(m)

        return m

    @pytest.fixture(scope="class")
    def iapws_turb(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = iapws95.Iapws95ParameterBlock()

        m.fs.unit = PressureChanger(default={
                "property_package": m.fs.properties,
                "thermodynamic_assumption": ThermodynamicAssumption.isentropic,
                "compressor": False})
        iscale.calculate_scaling_factors(m)
        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, iapws):
        assert len(iapws.fs.unit.inlet.vars) == 3
        assert hasattr(iapws.fs.unit.inlet, "flow_mol")
        assert hasattr(iapws.fs.unit.inlet, "enth_mol")
        assert hasattr(iapws.fs.unit.inlet, "pressure")

        assert hasattr(iapws.fs.unit, "outlet")
        assert len(iapws.fs.unit.outlet.vars) == 3
        assert hasattr(iapws.fs.unit.outlet, "flow_mol")
        assert hasattr(iapws.fs.unit.outlet, "enth_mol")
        assert hasattr(iapws.fs.unit.outlet, "pressure")

        assert hasattr(iapws.fs.unit, "work_mechanical")
        assert hasattr(iapws.fs.unit, "deltaP")
        assert isinstance(iapws.fs.unit.ratioP, Var)
        assert isinstance(iapws.fs.unit.ratioP_calculation, Constraint)

        assert isinstance(iapws.fs.unit.efficiency_isentropic, Var)
        assert isinstance(iapws.fs.unit.work_isentropic, Var)

        assert hasattr(iapws.fs.unit, "properties_isentropic")
        assert isinstance(iapws.fs.unit.isentropic_pressure, Constraint)
        assert isinstance(iapws.fs.unit.state_material_balances, Constraint)
        assert isinstance(iapws.fs.unit.isentropic, Constraint)
        assert isinstance(iapws.fs.unit.isentropic_energy_balance, Constraint)
        assert isinstance(iapws.fs.unit.actual_work, Constraint)

        assert number_variables(iapws) == 14
        assert number_total_constraints(iapws) == 9
        assert number_unused_variables(iapws) == 0

    @pytest.mark.component
    def test_units(self, iapws):
        assert_units_consistent(iapws)
        assert_units_equivalent(iapws.fs.unit.work_mechanical[0], units.W)
        assert_units_equivalent(iapws.fs.unit.deltaP[0], units.Pa)

    @pytest.mark.unit
    def test_dof(self, iapws):
        assert degrees_of_freedom(iapws) == 0

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, iapws):
        initialization_tester(iapws)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, iapws):
        results = solver.solve(iapws)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, iapws):
        # Check that outlet and isentropic pressure are equal
        assert pytest.approx(
            value(iapws.fs.unit.properties_isentropic[0].pressure), 1e-6) == \
            value(iapws.fs.unit.outlet.pressure[0])

        # Check that inlet and isentropic entropies are equal
        assert pytest.approx(
            value(iapws.fs.unit.properties_isentropic[0].entr_mol), 1e-6) == \
            value(iapws.fs.unit.control_volume.properties_in[0].entr_mol)

        assert pytest.approx(100, abs=1e-5) == \
            value(iapws.fs.unit.outlet.flow_mol[0])

        assert pytest.approx(4002, abs=1e0) == \
            value(iapws.fs.unit.outlet.enth_mol[0])

        assert pytest.approx(151325, abs=1e2) == \
            value(iapws.fs.unit.outlet.pressure[0])

        assert pytest.approx(101.43796915073504, abs=1e-1) == \
            value(iapws.fs.unit.work_mechanical[0])

        assert pytest.approx(91.29417223566153, abs=1e-1) == \
            value(iapws.fs.unit.work_isentropic[0])

        # For verification, check outlet and isentropic temperatures
        assert pytest.approx(326.170, 1e-5) == \
            value(iapws.fs.unit.control_volume.properties_out[0].temperature)

        assert pytest.approx(326.170, 1e-5) == \
            value(iapws.fs.unit.properties_isentropic[0].temperature)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, iapws):
        assert abs(value(iapws.fs.unit.inlet.flow_mol[0] -
                         iapws.fs.unit.outlet.flow_mol[0])) <= 1e-6

        assert abs(value(
                   iapws.fs.unit.outlet.flow_mol[0] *
                   (iapws.fs.unit.inlet.enth_mol[0] -
                    iapws.fs.unit.outlet.enth_mol[0]) +
                   iapws.fs.unit.work_mechanical[0])) <= 1e-6

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.integration
    def test_verify(self, iapws_turb):
        iapws = iapws_turb
        # Verify the turbine results against 3 known test cases
        # Case Data (90% isentropic efficency)
        cases = {
            "F": (1000, 1000, 1000),  # mol/s
            "Tin": (500, 800, 400),  # K
            "Pin": (1000, 10000, 200),  # kPa
            "W": (-1224.64, -1911.55, -1010.64),  # kW
            "Tout": (463.435, 742.992, 382.442),  # K
            "Pout": (700, 7000, 140),  # kPa
            "xout": (1.0, 1.0, 0.9886),  # vapor fraction
            "Tisen": (460.149, 738.224, 382.442),
        }

        for i in [0, 1, 2]:
            F = cases["F"][i]
            Tin = cases["Tin"][i]
            Tout = cases["Tout"][i]
            Pin = cases["Pin"][i]*1000
            Pout = cases["Pout"][i]*1000
            hin = iapws95.htpx(T=Tin*units.K, P=Pin*units.Pa)
            W = cases["W"][i]*1000
            Tis = cases["Tisen"][i]
            xout = cases["xout"][i]

            iapws.fs.unit.inlet.flow_mol[0].fix(F)
            iapws.fs.unit.inlet.enth_mol[0].fix(hin)
            iapws.fs.unit.inlet.pressure[0].fix(Pin)
            iapws.fs.unit.deltaP.fix(Pout - Pin)
            iapws.fs.unit.efficiency_isentropic.fix(0.9)
            iapws.fs.unit.initialize(optarg={'tol': 1e-6})
            results = solver.solve(iapws)
            # Check for optimal solution
            assert results.solver.termination_condition == \
                TerminationCondition.optimal
            assert results.solver.status == SolverStatus.ok

            Tout = pytest.approx(cases["Tout"][i], rel=1e-2)
            Pout = pytest.approx(cases["Pout"][i]*1000, rel=1e-2)
            Pout = pytest.approx(cases["Pout"][i]*1000, rel=1e-2)
            W = pytest.approx(cases["W"][i]*1000, rel=1e-2)
            xout = pytest.approx(xout, rel=1e-2)
            prop_out = iapws.fs.unit.control_volume.properties_out[0]
            prop_in = iapws.fs.unit.control_volume.properties_in[0]
            prop_is = iapws.fs.unit.properties_isentropic[0]

            assert value(prop_in.temperature) == pytest.approx(Tin, rel=1e-3)
            assert value(prop_is.temperature) == pytest.approx(Tis, rel=1e-3)
            assert value(iapws.fs.unit.control_volume.work[0]) == W
            assert value(prop_out.pressure) == Pout
            assert value(prop_out.temperature) == Tout
            assert value(prop_out.vapor_frac) == xout

    @pytest.mark.ui
    @pytest.mark.unit
    def test_report(self, iapws):
        iapws.fs.unit.report()


# -----------------------------------------------------------------------------
class TestSaponification(object):
    @pytest.fixture(scope="class")
    def sapon(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = SaponificationParameterBlock()

        m.fs.unit = PressureChanger(default={
                "property_package": m.fs.properties,
                "thermodynamic_assumption": ThermodynamicAssumption.pump,
                "compressor": False})

        m.fs.unit.inlet.flow_vol[0].fix(1e-3)
        m.fs.unit.inlet.temperature[0].fix(320)
        m.fs.unit.inlet.pressure[0].fix(101325)
        m.fs.unit.inlet.conc_mol_comp[0, "H2O"].fix(55388.0)
        m.fs.unit.inlet.conc_mol_comp[0, "NaOH"].fix(100.0)
        m.fs.unit.inlet.conc_mol_comp[0, "EthylAcetate"].fix(100.0)
        m.fs.unit.inlet.conc_mol_comp[0, "SodiumAcetate"].fix(0.0)
        m.fs.unit.inlet.conc_mol_comp[0, "Ethanol"].fix(0.0)

        m.fs.unit.deltaP.fix(-20000)
        m.fs.unit.efficiency_pump.fix(0.9)
        iscale.calculate_scaling_factors(m)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, sapon):
        assert len(sapon.fs.unit.inlet.vars) == 4
        assert hasattr(sapon.fs.unit.inlet, "flow_vol")
        assert hasattr(sapon.fs.unit.inlet, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.inlet, "temperature")
        assert hasattr(sapon.fs.unit.inlet, "pressure")

        assert len(sapon.fs.unit.outlet.vars) == 4
        assert hasattr(sapon.fs.unit.outlet, "flow_vol")
        assert hasattr(sapon.fs.unit.outlet, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.outlet, "temperature")
        assert hasattr(sapon.fs.unit.outlet, "pressure")

        assert hasattr(sapon.fs.unit, "work_mechanical")
        assert hasattr(sapon.fs.unit, "deltaP")
        assert isinstance(sapon.fs.unit.ratioP, Var)
        assert isinstance(sapon.fs.unit.ratioP_calculation, Constraint)

        assert isinstance(sapon.fs.unit.efficiency_pump, Var)
        assert isinstance(sapon.fs.unit.work_fluid, Var)
        assert isinstance(sapon.fs.unit.fluid_work_calculation, Constraint)
        assert isinstance(sapon.fs.unit.actual_work, Constraint)

        assert number_variables(sapon) == 21
        assert number_total_constraints(sapon) == 11
        assert number_unused_variables(sapon) == 0

    @pytest.mark.component
    def test_units(self, sapon):
        assert_units_consistent(sapon)
        assert_units_equivalent(sapon.fs.unit.work_mechanical[0], units.W)
        assert_units_equivalent(sapon.fs.unit.deltaP[0], units.Pa)

    @pytest.mark.unit
    def test_dof(self, sapon):
        assert degrees_of_freedom(sapon) == 0

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, sapon):
        initialization_tester(sapon)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, sapon):
        results = solver.solve(sapon)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, sapon):
        assert pytest.approx(1e-3, abs=1e-6) == \
            value(sapon.fs.unit.outlet.flow_vol[0])

        assert pytest.approx(55388.0, abs=1e-2) == value(
                sapon.fs.unit.outlet.conc_mol_comp[0, "H2O"])
        assert pytest.approx(100.0, abs=1e-2) == value(
                sapon.fs.unit.outlet.conc_mol_comp[0, "NaOH"])
        assert pytest.approx(100.0, abs=1e-2) == value(
                sapon.fs.unit.outlet.conc_mol_comp[0, "EthylAcetate"])
        assert pytest.approx(0.0, abs=1e-2) == value(
                sapon.fs.unit.outlet.conc_mol_comp[0, "SodiumAcetate"])
        assert pytest.approx(0.0, abs=1e-2) == value(
                sapon.fs.unit.outlet.conc_mol_comp[0, "Ethanol"])

        assert pytest.approx(320.0, abs=1e-1) == \
            value(sapon.fs.unit.outlet.temperature[0])

        assert pytest.approx(81325, abs=1e2) == \
            value(sapon.fs.unit.outlet.pressure[0])

        assert pytest.approx(-18.0, abs=1e-2) == \
            value(sapon.fs.unit.work_mechanical[0])
        assert pytest.approx(-20.0, abs=1e-2) == \
            value(sapon.fs.unit.work_fluid[0])

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, sapon):
        assert abs(value(
                sapon.fs.unit.outlet.flow_vol[0] *
                sapon.fs.properties.dens_mol*sapon.fs.properties.cp_mol *
                (sapon.fs.unit.inlet.temperature[0] -
                 sapon.fs.unit.outlet.temperature[0]) +
                sapon.fs.unit.work_mechanical[0])) <= 1e-4

    @pytest.mark.ui
    @pytest.mark.unit
    def test_report(self, sapon):
        sapon.fs.unit.report()


class TestTurbine(object):
    @pytest.mark.unit
    def test_config(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = PhysicalParameterTestBlock()

        m.fs.unit = Turbine(default={
                "property_package": m.fs.properties})
        iscale.calculate_scaling_factors(m)

        assert isinstance(m.fs.unit, PressureChangerData)
        # Check unit config arguments
        assert len(m.fs.unit.config) == 12

        assert m.fs.unit.config.material_balance_type == \
            MaterialBalanceType.useDefault
        assert m.fs.unit.config.energy_balance_type == \
            EnergyBalanceType.useDefault
        assert m.fs.unit.config.momentum_balance_type == \
            MomentumBalanceType.pressureTotal
        assert not m.fs.unit.config.has_phase_equilibrium
        assert not m.fs.unit.config.compressor
        assert m.fs.unit.config.thermodynamic_assumption == \
            ThermodynamicAssumption.isentropic
        assert m.fs.unit.config.property_package is m.fs.properties

        assert_units_consistent(m.fs.unit)


class TestCompressor(object):
    @pytest.mark.unit
    def test_config(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = PhysicalParameterTestBlock()

        m.fs.unit = Compressor(default={
                "property_package": m.fs.properties})
        iscale.calculate_scaling_factors(m)

        assert isinstance(m.fs.unit, PressureChangerData)
        # Check unit config arguments
        assert len(m.fs.unit.config) == 12

        assert m.fs.unit.config.material_balance_type == \
            MaterialBalanceType.useDefault
        assert m.fs.unit.config.energy_balance_type == \
            EnergyBalanceType.useDefault
        assert m.fs.unit.config.momentum_balance_type == \
            MomentumBalanceType.pressureTotal
        assert not m.fs.unit.config.has_phase_equilibrium
        assert m.fs.unit.config.compressor
        assert m.fs.unit.config.thermodynamic_assumption == \
            ThermodynamicAssumption.isentropic
        assert m.fs.unit.config.property_package is m.fs.properties

        assert_units_consistent(m.fs.unit)


class TestPump(object):
    @pytest.mark.unit
    def test_config(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = PhysicalParameterTestBlock()

        m.fs.unit = Pump(default={
                "property_package": m.fs.properties})
        iscale.calculate_scaling_factors(m)

        assert isinstance(m.fs.unit, PressureChangerData)
        # Check unit config arguments
        assert len(m.fs.unit.config) == 12

        assert m.fs.unit.config.material_balance_type == \
            MaterialBalanceType.useDefault
        assert m.fs.unit.config.energy_balance_type == \
            EnergyBalanceType.useDefault
        assert m.fs.unit.config.momentum_balance_type == \
            MomentumBalanceType.pressureTotal
        assert not m.fs.unit.config.has_phase_equilibrium
        assert m.fs.unit.config.compressor
        assert m.fs.unit.config.thermodynamic_assumption == \
            ThermodynamicAssumption.pump
        assert m.fs.unit.config.property_package is m.fs.properties

        assert_units_consistent(m.fs.unit)


@pytest.mark.skipif(not iapws95.iapws95_available(),
                    reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
class Test_costing(object):
    @pytest.mark.component
    def test_pump(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.properties = iapws95.Iapws95ParameterBlock()
        m.fs.unit = PressureChanger(default={
                "property_package": m.fs.properties,
                "thermodynamic_assumption": ThermodynamicAssumption.pump,
                "compressor": True})
        # set inputs
        m.fs.unit.inlet.flow_mol[0].fix(10000)
        m.fs.unit.inlet.enth_mol[0].fix(4000)
        m.fs.unit.inlet.pressure[0].fix(101325)
        m.fs.unit.deltaP.fix(50000)
        m.fs.unit.efficiency_pump.fix(0.9)
        iscale.calculate_scaling_factors(m)

        assert degrees_of_freedom(m) == 0

        m.fs.unit.get_costing(pump_type='centrifugal',
                              Mat_factor='nickel',
                              pump_motor_type_factor='enclosed')

        m.fs.unit.initialize()
        # check costing block initialization
        assert m.fs.unit.costing.purchase_cost.value == \
            pytest.approx(70115.019, abs=1e-2)

        assert_units_consistent(m.fs.unit)

        solver.solve(m, tee=True)
        assert m.fs.unit.costing.purchase_cost.value == \
            pytest.approx(70115.0, 1e-5)

    @pytest.mark.component
    def test_compressor(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.properties = iapws95.Iapws95ParameterBlock()
        m.fs.unit = PressureChanger(default={
                "property_package": m.fs.properties,
                "thermodynamic_assumption":
                ThermodynamicAssumption.isentropic,
                "compressor": True})
        # set inputs
        m.fs.unit.inlet.flow_mol[0].fix(10000)
        m.fs.unit.inlet.enth_mol[0].fix(4000)
        m.fs.unit.inlet.pressure[0].fix(101325)
        m.fs.unit.deltaP.fix(500000)
        m.fs.unit.efficiency_isentropic.fix(0.9)
        iscale.set_scaling_factor(m.fs.unit.control_volume.work[0], 1e-5)
        iscale.calculate_scaling_factors(m)

        assert degrees_of_freedom(m) == 0
        m.fs.unit.get_costing(mover_type="compressor")
        m.fs.unit.initialize()
        results = solver.solve(m)
        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

        assert value(m.fs.unit.control_volume.work[0]) == \
            pytest.approx(101410.4, rel=1e-5)

        assert m.fs.unit.costing.purchase_cost.value == \
            pytest.approx(334598, rel=1e-5)

        assert_units_consistent(m.fs.unit)

        solver.solve(m, tee=True)
        assert m.fs.unit.costing.purchase_cost.value == \
            pytest.approx(334598, rel=1e-5)

    @pytest.mark.component
    def test_turbine(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.properties = iapws95.Iapws95ParameterBlock()
        m.fs.unit = PressureChanger(default={
                "property_package": m.fs.properties,
                "thermodynamic_assumption":
                ThermodynamicAssumption.isentropic,
                "compressor": False})
        # set inputs
        m.fs.unit.inlet.flow_mol[0].fix(1000)  # mol/s
        Tin = 500  # K
        Pin = 1000000  # Pa
        Pout = 700000  # Pa
        hin = iapws95.htpx(Tin*units.K, Pin*units.Pa)
        m.fs.unit.inlet.enth_mol[0].fix(hin)
        m.fs.unit.inlet.pressure[0].fix(Pin)

        m.fs.unit.deltaP.fix(Pout - Pin)
        m.fs.unit.efficiency_isentropic.fix(0.9)
        # build costing block before initializing the unit
        m.fs.unit.get_costing()

        m.fs.unit.initialize()
        # check costing initialization is working
        assert m.fs.unit.costing.purchase_cost.value ==\
            pytest.approx(213199, 1e-5)

        assert degrees_of_freedom(m) == 0

        assert_units_consistent(m.fs.unit)

        solver.solve(m, tee=True)
        assert m.fs.unit.costing.purchase_cost.value ==\
            pytest.approx(213199, 1e-5)

    @pytest.mark.component
    def test_turbine_performance_way1(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.properties = iapws95.Iapws95ParameterBlock()

        def perf_callback(b):
            unit_hd = units.J/units.kg
            unit_vflw = units.m**3/units.s
            @b.Constraint(m.fs.config.time)
            def pc_isen_eff_eqn(b, t):
                prnt = b.parent_block()
                vflw = prnt.control_volume.properties_in[t].flow_vol
                return prnt.efficiency_isentropic[t] == 0.9/3.975*vflw/unit_vflw
            @b.Constraint(m.fs.config.time)
            def pc_isen_head_eqn(b, t):
                prnt = b.parent_block()
                vflw = prnt.control_volume.properties_in[t].flow_vol
                return b.head_isentropic[t]/1000 == \
                    -75530.8/3.975/1000*vflw/unit_vflw*unit_hd

        m.fs.unit = Turbine(default={
            "property_package": m.fs.properties,
            "support_isentropic_performance_curves":True,
            "isentropic_performance_curves": {"build_callback": perf_callback}})

        # set inputs
        m.fs.unit.inlet.flow_mol[0].fix(1000)  # mol/s
        Tin = 500  # K
        Pin = 1000000  # Pa
        Pout = 700000  # Pa
        hin = iapws95.htpx(Tin*units.K, Pin*units.Pa)
        m.fs.unit.inlet.enth_mol[0].fix(hin)
        m.fs.unit.inlet.pressure[0].fix(Pin)

        m.fs.unit.initialize()
        assert degrees_of_freedom(m) == 0
        assert_units_consistent(m.fs.unit)
        solver.solve(m, tee=True)

        assert value(m.fs.unit.efficiency_isentropic[0]) \
            == pytest.approx(0.9, rel=1e-3)
        assert value(m.fs.unit.deltaP[0]) == pytest.approx(-3e5, rel=1e-3)

    @pytest.mark.component
    def test_turbine_performance_way2(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.properties = iapws95.Iapws95ParameterBlock()
        m.fs.unit = Turbine(default={
            "property_package": m.fs.properties,
            "support_isentropic_performance_curves":True})
        unit_hd = units.J/units.kg
        unit_vflw = units.m**3/units.s
        @m.fs.unit.performance_curve.Constraint(m.fs.config.time)
        def pc_isen_eff_eqn(b, t):
            prnt = b.parent_block()
            vflw = prnt.control_volume.properties_in[t].flow_vol
            return prnt.efficiency_isentropic[t] == 0.9/3.975*vflw/unit_vflw
        @m.fs.unit.performance_curve.Constraint(m.fs.config.time)
        def pc_isen_head_eqn(b, t):
            prnt = b.parent_block()
            vflw = prnt.control_volume.properties_in[t].flow_vol
            return b.head_isentropic[t]/1000 == \
                -75530.8/3.975/1000*vflw/unit_vflw*unit_hd

        # set inputs
        m.fs.unit.inlet.flow_mol[0].fix(1000)  # mol/s
        Tin = 500  # K
        Pin = 1000000  # Pa
        Pout = 700000  # Pa
        hin = iapws95.htpx(Tin*units.K, Pin*units.Pa)
        m.fs.unit.inlet.enth_mol[0].fix(hin)
        m.fs.unit.inlet.pressure[0].fix(Pin)

        m.fs.unit.initialize()
        assert degrees_of_freedom(m) == 0
        assert_units_consistent(m.fs.unit)
        solver.solve(m, tee=True)

        assert value(m.fs.unit.efficiency_isentropic[0]) \
            == pytest.approx(0.9, rel=1e-3)
        assert value(m.fs.unit.deltaP[0]) == pytest.approx(-3e5, rel=1e-3)



from idaes.gas_solid_contactors.properties.oxygen_iron_OC_oxidation. \
      gas_phase_thermo import (GasPhaseParameterBlock,
                               GasPhaseStateBlock)
from idaes.gas_solid_contactors.properties.oxygen_iron_OC_oxidation. \
    solid_phase_thermo import (SolidPhaseParameterBlock,
                               SolidPhaseStateBlock)
from idaes.gas_solid_contactors.properties.oxygen_iron_OC_oxidation. \
    hetero_reactions import (HeteroReactionParameterBlock,
                             ReactionBlock)
